from __future__ import annotations

# standard libraries
import abc
import contextlib
import copy
import datetime
import functools
import gettext
import operator
import pathlib
import threading
import types
import typing
import uuid
import warnings
import weakref

# third party libraries
import numpy
import numpy.typing

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.data import Image
from nion.swift.model import ApplicationData
from nion.swift.model import DynamicString
from nion.swift.model import Metadata
from nion.swift.model import Persistence
from nion.swift.model import Utility
from nion.utils import DateTime
from nion.utils import Event
from nion.utils import Stream
from nion.utils import ReferenceCounting

if typing.TYPE_CHECKING:
    from nion.swift.model import DisplayItem
    from nion.swift.model import DocumentModel
    from nion.swift.model import Project

from nion.data.DataAndMetadata import _ImageDataType

_ = gettext.gettext

UNTITLED_STR = _("Untitled")


class CalibrationList:

    def __init__(self, calibrations: typing.Optional[DataAndMetadata.CalibrationListType] = None) -> None:
        self.list = list() if calibrations is None else copy.deepcopy(list(calibrations))

    def read_dict(self, storage_list: typing.Sequence[typing.Mapping[str, typing.Any]]) -> CalibrationList:
        # storage_list will be whatever is returned by write_dict.
        new_list: typing.List[Calibration.Calibration] = list()
        for calibration_dict in storage_list:
            new_list.append(Calibration.Calibration().read_dict(calibration_dict))
        self.list = new_list
        return self  # for convenience

    def write_dict(self) -> typing.List[Persistence.PersistentDictType]:
        l: typing.List[Persistence.PersistentDictType] = list()
        for calibration in self.list:
            l.append(calibration.write_dict())
        return l


"""
    Data sources are interfaces to get data and metadata.

    Data sources should support deep copy. They are never shared (i.e. they only occur once in the model).

    When data sources is about to be closed, they will receive this message.

    closed()

    When data sources are about to be removed, they will receive this message.

    about_to_be_removed()

    Data sources can only be associated in the operation tree of a single data item. The data source
    keeps track of that so as to maintain the dependent_data_item list on data items.

    set_dependent_data_item(data_item)

    Data sources can subscribe to other data items becoming available via the data item manager.
"""


class DtypeToStringConverter:
    def convert(self, value: typing.Optional[numpy.typing.DTypeLike]) -> typing.Optional[str]:
        return str(value) if value is not None else None

    def convert_back(self, value: typing.Optional[str]) -> typing.Optional[numpy.typing.DTypeLike]:
        return numpy.dtype(value) if value is not None else None


class DatetimeToStringConverter:
    def convert(self, value: typing.Optional[datetime.datetime]) -> typing.Optional[str]:
        return value.isoformat() if value is not None else None

    def convert_back(self, value: typing.Optional[str]) -> typing.Optional[datetime.datetime]:
        try:
            if value and len(value) == 26:
                return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%f")
            elif value and len(value) == 19:
                return datetime.datetime.strptime(value, "%Y-%m-%dT%H:%M:%S")
        except ValueError as e:
            pass  # fall through
        return None


class UuidsToStringsConverter:
    def convert(self, value: typing.List[uuid.UUID]) -> typing.List[str]:
        return [str(uuid_) for uuid_ in value]

    def convert_back(self, value: typing.List[str]) -> typing.List[uuid.UUID]:
        return [uuid.UUID(uuid_str) for uuid_str in value]


class SessionManager(abc.ABC):

    @property
    @abc.abstractmethod
    def current_session_id(self) -> typing.Optional[str]:
        pass


# dates are _local_ time and must use this specific ISO 8601 format. 2013-11-17T08:43:21.389391
# time zones are offsets (east of UTC) in the following format "+HHMM" or "-HHMM"
# daylight savings times are time offset (east of UTC) in format "+MM" or "-MM"
# time zone name is for display only and has no specified format

class DataItem(Persistence.PersistentObject):
    """
    Data items represent a data, description, display, and graphics within a library.

    * *created* a datetime item
    * *session_id* a string representing the session

    *Descriptive Metadata*

    * *title* a string (single line)
    * *caption* a string (multiple lines)
    * *rating* an integer star rating (0 to 5)
    * *flag* a flag (-1, 0, 1)

    *Modification Dates*

    Data items keep track of a created datetime, which is set once when the object is created; a modified datetime,
    which is updated whenever the object or its related objects get modified.

    Data items also provide a data modified datetime, which is the latest data modified datetime from all of the data
    sources.

    *Notifications*

    Data items will emit the following notifications to listeners. Listeners should take care to not call functions
    which result in cycles of notifications. For instance, functions handling data_item_content_changed should not use
    functions that will trigger the data to be computed which will emit data_item_content_changed, resulting in a cycle.

    * data_item_content_changed(data_item, changes)

    data_item_content_changed is invoked when the content of the data item changes. The changes parameter is a set of
    changes from DATA, METADATA, DISPLAYS, SOURCE. This may be called on a thread.

    *Miscellaneous*

    Transactions.

    Live state.

    Snapshots and deep copies.

    Properties.

    Metadata Note: Until data source is merged into data item, metadata access cannot be done in this base class;
    instead it is implemented in the derived classes. It is a regular field in the composite library item; but in
    the data item, it is stored in the data source. When data source is merged, it will become a regular field
    and can be moved into this class.
    """

    storage_version = 13
    writer_version = 13

    def __init__(self, data: typing.Optional[_ImageDataType] = None, item_uuid: typing.Optional[uuid.UUID] = None,
                 large_format: bool = False) -> None:
        super().__init__()
        self.uuid = item_uuid if item_uuid else self.uuid
        self.large_format = large_format
        self._document_model: typing.Optional[DocumentModel.DocumentModel] = None  # used only for Facade
        self.define_type("data-item")
        self.define_property("created", DateTime.utcnow(), hidden=True, converter=DatetimeToStringConverter(), changed=self.__description_property_changed)
        data_shape = data.shape if data is not None else None
        data_dtype = data.dtype if data is not None else None
        dimensional_shape = Image.dimensional_shape_from_shape_and_dtype(data_shape, data_dtype)
        collection_dimension_count = (2 if len(dimensional_shape) == 3 else 0) if dimensional_shape is not None else None
        datum_dimension_count = len(dimensional_shape) - collection_dimension_count if dimensional_shape is not None and collection_dimension_count is not None else None
        self.define_property("data_shape", data_shape, hidden=True, recordable=False, is_equal_fn=operator.eq, changed=self.__property_changed)
        self.define_property("data_dtype", data_dtype, hidden=True, recordable=False, converter=DtypeToStringConverter(), changed=self.__property_changed, is_equal_fn=operator.eq)
        self.define_property("is_sequence", False, hidden=True, recordable=False, changed=self.__data_description_changed)
        self.define_property("collection_dimension_count", collection_dimension_count, hidden=True, recordable=False, changed=self.__data_description_changed, value_type=int)
        self.define_property("datum_dimension_count", datum_dimension_count, hidden=True, recordable=False, changed=self.__data_description_changed, value_type=int)
        self.define_property("intensity_calibration", Calibration.Calibration(), hidden=True, make=Calibration.Calibration, changed=self.__metadata_property_changed, is_equal_fn=operator.eq)
        self.define_property("dimensional_calibrations", CalibrationList(), hidden=True, make=CalibrationList, changed=self.__dimensional_calibrations_changed, is_equal_fn=operator.eq)
        self.define_property("data_modified", hidden=True, recordable=False, converter=DatetimeToStringConverter(), changed=self.__metadata_property_changed, value_type=datetime.datetime)
        self.define_property("timezone", Utility.get_local_timezone(), hidden=True, changed=self.__timezone_property_changed, recordable=False, value_type=str)
        self.define_property("timezone_offset", Utility.TimezoneMinutesToStringConverter().convert(Utility.local_utcoffset_minutes()), hidden=True, changed=self.__timezone_property_changed, recordable=False, value_type=str)
        self.define_property("metadata", dict(), hidden=True, changed=self.__metadata_property_changed, is_equal_fn=Utility.deep_compare_items)
        self.define_property("title", str(), changed=self.__property_changed, hidden=True)
        self.define_property("dynamic_title", hidden=True)
        self.define_property("dynamic_title_enabled", True, changed=self.__property_changed, hidden=True)
        self.define_property("caption", changed=self.__property_changed, hidden=True, value_type=str)
        self.define_property("description", changed=self.__property_changed, hidden=True, value_type=str)
        self.define_property("source_specifier", changed=self.__source_specifier_changed, key="source_uuid", hidden=True)
        self.define_property("session_id", validate=self.__validate_session_id, changed=self.__property_changed, hidden=True, value_type=str)
        self.define_property("session", dict(), changed=self.__property_changed, hidden=True)
        self.define_property("category", "persistent", changed=self.__property_changed, hidden=True, value_type=str)
        self.__data: typing.Optional[_ImageDataType] = None
        self.__data_metadata: typing.Optional[DataAndMetadata.DataMetadata] = None
        self.__data_and_metadata_unloadable = False
        self.__data_and_metadata_first_update_after_reserve = False
        self.__data_and_metadata_lock = threading.RLock()
        self.__intensity_calibration: typing.Optional[Calibration.Calibration] = None
        self.__dimensional_calibrations: typing.List[Calibration.Calibration] = list()
        self.__metadata: typing.Dict[str, typing.Any] = dict()
        self.__data_ref_count = 0
        self.__data_ref_count_mutex = threading.RLock()
        self.__pending_write = True
        self.__in_transaction_state = False
        self.__write_delay_modified_count = 0
        self.__write_delay_data_changed = False
        self.__source_file_path: typing.Optional[pathlib.Path] = None
        self.__is_live = False
        self.__session_manager: typing.Optional[SessionManager] = None
        self.__source_reference = self.create_item_reference()
        self.description_changed_event = Event.Event()
        self.item_changed_event = Event.Event()
        self.data_item_changed_event = Event.Event()  # anything has changed
        self.data_changed_event = Event.Event()  # data has changed
        self.will_change_event = Event.Event()
        self.did_change_event = Event.Event()
        self.__data_item_change_count = 0
        self.__data_item_change_count_lock = threading.RLock()
        self.__change_thread: typing.Optional[threading.Thread] = None
        self.__change_count = 0
        self.__change_count_lock = threading.RLock()
        self.__change_changed = False
        self.__change_data_changed = False
        self.__pending_xdata_lock = threading.RLock()
        self.__pending_xdata: typing.Optional[DataAndMetadata.DataAndMetadata] = None
        self.__pending_queue: typing.List[typing.Tuple[DataAndMetadata.DataAndMetadata, typing.Sequence[slice], typing.Sequence[slice], DataAndMetadata.DataMetadata]] = list()
        self.__content_changed = False
        self.__display_data_channel_refs = set[weakref.ReferenceType["DisplayItem.DisplayDataChannel"]]()
        if data is not None:
            data_and_metadata = DataAndMetadata.new_data_and_metadata(data, timezone=self.timezone, timezone_offset=self.timezone_offset)
            self.increment_data_ref_count()
            try:
                self.__set_data_and_metadata_direct(data_and_metadata)
            finally:
                self.decrement_data_ref_count()
        self.__dynamic_title: typing.Optional[DynamicString.DynamicString] = None
        self.__dynamic_title_persistence_action: typing.Optional[Stream.ValueStreamAction[Persistence.PersistentDictType]] = None
        self.__specified_title_stream = Stream.ValueStream[str]()
        self.__source_title_stream = Stream.FollowStream[str]()
        self.__computation_title_stream = Stream.ValueStream[str]()
        self.__dynamic_title_stream = Stream.FollowStream[str]()
        self.__dynamic_title_enabled_stream = Stream.ValueStream[bool](True)

        def combine_display_title(
                specified_title: typing.Optional[str],
                source_title: typing.Optional[str],
                computation_title: typing.Optional[str],
                dynamic_title: typing.Optional[str],
                dynamic_title_enabled: typing.Optional[bool]
        ) -> str:
            if dynamic_title_enabled:
                if dynamic_title:
                    return dynamic_title
            if dynamic_title_enabled or not specified_title:
                if source_title and computation_title:
                    return f"{source_title} ({computation_title})"
            if specified_title:
                return specified_title
            return UNTITLED_STR

        self.__title_stream = Stream.CombineLatestStream([self.__specified_title_stream, self.__source_title_stream, self.__computation_title_stream, self.__dynamic_title_stream, self.__dynamic_title_enabled_stream], combine_display_title)

        def update_dynamic_title(data_item: DataItem, dynamic_title_str: typing.Optional[str]) -> None:
            if not data_item._is_reading:
                data_item._set_persistent_property_value("title", dynamic_title_str)

        self.__title_stream_action = Stream.ValueStreamAction(self.__title_stream, ReferenceCounting.weak_partial(update_dynamic_title, self))

        def combine_placeholder_title(
                source_title: typing.Optional[str],
                computation_title: typing.Optional[str],
                dynamic_title: typing.Optional[str],
        ) -> str:
            if dynamic_title:
                return dynamic_title
            if source_title and computation_title:
                return f"{source_title} ({computation_title})"
            return str()

        self.__placeholder_title_stream = Stream.CombineLatestStream([self.__source_title_stream, self.__computation_title_stream, self.__dynamic_title_stream], combine_placeholder_title)

    def close(self) -> None:
        self.__title_stream_action = typing.cast(typing.Any, None)
        self.__title_stream = typing.cast(typing.Any, None)
        self.__dynamic_title = None
        self.__dynamic_title_persistence_action = None
        self.__specified_title_stream = typing.cast(typing.Any, None)
        self.__source_title_stream = typing.cast(typing.Any, None)
        self.__computation_title_stream = typing.cast(typing.Any, None)
        self.__dynamic_title_stream = typing.cast(typing.Any, None)
        self.__dynamic_title_enabled_stream = typing.cast(typing.Any, None)
        self.__placeholder_title_stream = typing.cast(typing.Any, None)
        self.__data = None
        super().close()

    def __str__(self) -> str:
        return "{0} {1} ({2}, {3})".format(self.__repr__(), (self.title if self.title else _("Untitled")), str(self.uuid), self.date_for_sorting_local_as_string)

    def __copy__(self) -> DataItem:
        return copy.deepcopy(self)

    def __deepcopy__(self, memo: typing.Dict[typing.Any, typing.Any]) -> DataItem:
        data_item_copy = self.__class__()
        try:
            # data format (temporary until moved to buffered data source)
            data_item_copy.large_format = self.large_format
            # metadata
            data_item_copy.created = self.created
            data_item_copy.timezone = self.timezone
            data_item_copy.timezone_offset = self.timezone_offset
            data_item_copy.metadata = self.metadata
            data_item_copy.title = self.title
            data_item_copy.dynamic_title_enabled = self.dynamic_title_enabled
            data_item_copy.caption = self.caption
            data_item_copy.description = self.description
            data_item_copy.session_id = self.session_id
            data_item_copy.session_data = copy.deepcopy(self.session_data)
            data_item_copy.category = self.category
            # data and metadata
            data_item_copy.set_data_and_metadata(copy.deepcopy(self.data_and_metadata), self.data_modified)
            # copy this last and avoid making an extra unnecessary copy
            data_item_copy.__set_dynamic_title(copy.deepcopy(self.__dynamic_title) if self.__dynamic_title else None)
            memo[id(self)] = data_item_copy
            return data_item_copy
        except Exception:
            data_item_copy.close()
            raise

    def snapshot(self) -> DataItem:
        """Return a new library item which is a copy of this one with any dynamic behavior made static."""
        data_item_snapshot = copy.deepcopy(self)
        data_item_snapshot.category = "persistent"
        return data_item_snapshot

    def clone(self) -> DataItem:
        data_item = self.__class__()
        data_item.uuid = self.uuid
        return data_item

    @property
    def created(self) -> datetime.datetime:
        return typing.cast(datetime.datetime, self._get_persistent_property_value("created"))

    @created.setter
    def created(self, value: datetime.datetime) -> None:
        self._set_persistent_property_value("created", value)

    @ property
    def dynamic_title_enabled(self) -> bool:
        return typing.cast(bool, self._get_persistent_property_value("dynamic_title_enabled"))

    @dynamic_title_enabled.setter
    def dynamic_title_enabled(self, value: bool) -> None:
        self._set_persistent_property_value("dynamic_title_enabled", value)
        if self.dynamic_title_enabled:
            self.__set_dynamic_title(self.__dynamic_title)

    @property
    def title(self) -> str:
        return self.__title_stream.value or str()

    @title.setter
    def title(self, value: typing.Optional[str]) -> None:
        if value:
            self._set_persistent_property_value("dynamic_title_enabled", False)
            self._set_persistent_property_value("title", value)
        else:
            self._set_persistent_property_value("dynamic_title_enabled", True)

    @property
    def title_stream(self) -> Stream.AbstractStream[str]:
        return self.__title_stream

    @property
    def placeholder_title_stream(self) -> Stream.AbstractStream[str]:
        return self.__placeholder_title_stream

    @property
    def caption(self) -> typing.Optional[str]:
        return typing.cast(typing.Optional[str], self._get_persistent_property_value("caption"))

    @caption.setter
    def caption(self, value: typing.Optional[str]) -> None:
        self._set_persistent_property_value("caption", value)

    @property
    def description(self) -> typing.Optional[str]:
        return typing.cast(typing.Optional[str], self._get_persistent_property_value("description"))

    @description.setter
    def description(self, value: typing.Optional[str]) -> None:
        self._set_persistent_property_value("description", value)

    @property
    def category(self) -> str:
        return typing.cast(str, self._get_persistent_property_value("category"))

    @category.setter
    def category(self, value: str) -> None:
        self._set_persistent_property_value("category", value)

    @property
    def session_id(self) -> typing.Optional[str]:
        return typing.cast(typing.Optional[str], self._get_persistent_property_value("session_id"))

    @session_id.setter
    def session_id(self, value: typing.Optional[str]) -> None:
        assert value is None or datetime.datetime.strptime(value, "%Y%m%d-%H%M%S")
        self._set_persistent_property_value("session_id", value)

    @property
    def session(self) -> Persistence.PersistentDictType:
        return typing.cast(Persistence.PersistentDictType, self._get_persistent_property_value("session"))

    @session.setter
    def session(self, value: Persistence.PersistentDictType) -> None:
        self._set_persistent_property_value("session", value)

    @property
    def source_specifier(self) -> typing.Optional[Persistence._SpecifierType]:
        return typing.cast(typing.Optional[Persistence._SpecifierType], self._get_persistent_property_value("source_specifier"))

    @source_specifier.setter
    def source_specifier(self, value: typing.Optional[Persistence._SpecifierType]) -> None:
        self._set_persistent_property_value("source_specifier", value)

    @property
    def project(self) -> Project.Project:
        return typing.cast("Project.Project", self.container)

    def create_proxy(self) -> Persistence.PersistentObjectProxy[DataItem]:
        return self.project.create_item_proxy(item=self)

    @property
    def item_specifier(self) -> Persistence.PersistentObjectSpecifier:
        return Persistence.PersistentObjectSpecifier(self.uuid)

    def add_display_data_channel(self, display_data_channel: DisplayItem.DisplayDataChannel) -> None:
        """Add a display data channel referencing this data item.

        This is used as an optimization to be able to easily access the display data channels for a data item.
        Think of it as an "index" in the database sense.
        """
        self.__display_data_channel_refs.add(weakref.ref(display_data_channel))
        self.notify_add_item("display_data_channels", display_data_channel)

    def remove_display_data_channel(self, display_data_channel: DisplayItem.DisplayDataChannel) -> None:
        """Remove a display data channel referencing this data item."""
        self.__display_data_channel_refs.remove(weakref.ref(display_data_channel))
        self.notify_discard_item("display_data_channels", display_data_channel)

    @property
    def display_data_channels(self) -> typing.Set[DisplayItem.DisplayDataChannel]:
        """Return the list of display data channels referencing this data item."""
        display_data_channels = set["DisplayItem.DisplayDataChannel"]()
        for display_data_channel_ref in self.__display_data_channel_refs:
            display_data_channel = display_data_channel_ref()
            assert display_data_channel  # should never be None
            display_data_channels.add(display_data_channel)
        return display_data_channels

    @property
    def in_transaction_state(self) -> bool:
        return self.__in_transaction_state

    def __enter_write_delay_state(self) -> None:
        self.__write_delay_modified_count = self.modified_count
        self.__write_delay_data_changed = False
        if self.persistent_object_context:
            self.enter_write_delay()

    def __exit_write_delay_state(self) -> None:
        if self.persistent_object_context:
            self.exit_write_delay()
            self.__data_and_metadata_unloadable = True
            self._finish_pending_write()

    def write_to_dict(self) -> Persistence.PersistentDictType:
        properties = super().write_to_dict()
        properties["version"] = DataItem.writer_version
        return properties

    def write_data_if_not_delayed(self) -> None:
        if not self.is_write_delayed:
            # write the uuid and version explicitly
            self.property_changed("uuid", str(self.uuid))
            self.property_changed("version", DataItem.writer_version)
            self.__write_data()
            self.__pending_write = False

    def __write_data(self) -> None:
        if self.__data is not None:
            self.write_external_data("data", self.__data)

    def _finish_pending_write(self) -> None:
        if self.__pending_write:
            self.write_data_if_not_delayed()
        else:
            if self.modified_count > self.__write_delay_modified_count:
                self.rewrite()
            if self.__write_delay_data_changed:
                self.__write_data()

    def _transaction_state_entered(self) -> None:
        self.__in_transaction_state = True
        # first enter the write delay state.
        self.__enter_write_delay_state()
        # load data to prevent paging in and out.
        self.increment_data_ref_count()

    def _transaction_state_exited(self) -> None:
        self.__in_transaction_state = False
        # exit the write delay state.
        self.__exit_write_delay_state()
        # unload data.
        self.decrement_data_ref_count()

    @property
    def source(self) -> typing.Optional[Persistence.PersistentObject]:
        return self.__source_reference.item

    @source.setter
    def source(self, source: typing.Optional[Persistence.PersistentObject]) -> None:
        self.__source_reference.item = source
        self.source_specifier = Persistence.write_persistent_specifier(source.uuid) if source else None

    def __source_specifier_changed(self, name: str, d: Persistence._SpecifierType) -> None:
        self.__source_reference.item_specifier = Persistence.read_persistent_specifier(d)

    def source_data_items_changed(self, source_data_items: typing.List[DataItem]) -> None:
        if len(source_data_items) == 1:
            self.__source_title_stream.stream = source_data_items[0].title_stream
        elif len(source_data_items) > 1:
            def combine_source_titles(count: int, title: str | None) -> str:
                return f"{title} +{count - 1} More"

            self.__source_title_stream.stream = Stream.MapStream(source_data_items[0].title_stream, functools.partial(combine_source_titles, len(source_data_items)))
        else:
            self.__source_title_stream.stream = None

    def computation_title_changed(self, computation_title: typing.Optional[str]) -> None:
        self.__computation_title_stream.value = computation_title

    def persistent_object_context_changed(self) -> None:
        # handle case where persistent object context is set on an item that is already under transaction.
        # this can occur during acquisition. any other cases?
        super().persistent_object_context_changed()

        if self.__in_transaction_state:
            self.__enter_write_delay_state()
        else:
            self.__data_and_metadata_unloadable = self.persistent_object_context is not None and not self.is_write_delayed

    def __get_file_path(self) -> str:
        return typing.cast(str, getattr(self.persistent_storage, "get_storage_property")(self, "file_path")) if self.persistent_storage else str()

    def _test_get_file_path(self) -> str:
        # hack for test function. not a public method.
        return self.__get_file_path()

    def read_from_dict(self, properties: Persistence.PersistentDictType) -> None:
        self.large_format = properties.get("__large_format", self.large_format)
        # when reading, handle changes specially. first, put everything into a change
        # block; then make sure that no change notifications actually occur. this makes
        # sure things like cached values are preserved after reading.
        with self.data_item_changes():
            super().read_from_dict(properties)
            data_shape = self._get_persistent_property_value("data_shape", None)
            data_dtype = DtypeToStringConverter().convert_back(self._get_persistent_property_value("data_dtype", None))
            if data_shape is not None and data_dtype is not None:
                data_shape_and_dtype = data_shape, data_dtype
                dimensional_shape = Image.dimensional_shape_from_shape_and_dtype(data_shape, data_dtype)
                intensity_calibration = self._get_persistent_property_value("intensity_calibration")
                dimensional_calibration_list = self._get_persistent_property_value("dimensional_calibrations")
                dimensional_calibrations = dimensional_calibration_list.list if dimensional_calibration_list else None
                if dimensional_calibrations is not None:
                    assert dimensional_shape is not None
                    while len(dimensional_shape) > len(dimensional_calibrations):
                        dimensional_calibrations.append(Calibration.Calibration())
                    while len(dimensional_shape) < len(dimensional_calibrations):
                        dimensional_calibrations.pop(-1)
                metadata = self._get_persistent_property_value("metadata")
                timestamp = self._get_persistent_property_value("data_modified")
                if timestamp is None:  # invalid timestamp -- set property to now but don't trigger change
                    timestamp = self.created or DateTime.utcnow()
                    self._get_persistent_property("data_modified").value = timestamp
                is_sequence = self._get_persistent_property_value("is_sequence", False)
                collection_dimension_count = self._get_persistent_property_value("collection_dimension_count")
                datum_dimension_count = self._get_persistent_property_value("datum_dimension_count")
                if collection_dimension_count is None:
                    collection_dimension_count = 2 if dimensional_shape is not None and len(dimensional_shape) == 3 and not is_sequence else 0
                    # update collection_dimension_count, but in a way that sets the internal value but
                    # doesn't trigger a write to disk or a change modification.
                    self._get_persistent_property("collection_dimension_count").set_value(collection_dimension_count)
                if datum_dimension_count is None:
                    datum_dimension_count = len(dimensional_shape) - collection_dimension_count - (1 if is_sequence else 0) if dimensional_shape is not None and collection_dimension_count is not None else 0
                    # update collection_dimension_count, but in a way that sets the internal value but
                    # doesn't trigger a write to disk or a change modification.
                    self._get_persistent_property("datum_dimension_count").set_value(datum_dimension_count)
                data_descriptor = DataAndMetadata.DataDescriptor(is_sequence, collection_dimension_count, datum_dimension_count)
                self.__data_metadata = DataAndMetadata.DataMetadata(data_shape_and_dtype=data_shape_and_dtype,
                                                                    intensity_calibration=intensity_calibration,
                                                                    dimensional_calibrations=dimensional_calibrations,
                                                                    metadata=metadata,
                                                                    timestamp=timestamp,
                                                                    data_descriptor=data_descriptor,
                                                                    timezone=self.timezone,
                                                                    timezone_offset=self.timezone_offset)
                with self.__data_ref_count_mutex:
                    if self.__data_ref_count:
                        self.__load_data()
                self.__data_and_metadata_unloadable = self.persistent_object_context is not None
            else:
                metadata = self._get_persistent_property_value("metadata")
                self.__metadata = copy.deepcopy(metadata) if metadata else dict()
            self.__pending_write = False
            if self.created is None:  # invalid timestamp -- set property to now but don't trigger change
                self._get_persistent_property("created").value = DateTime.utcnow()
            self.__content_changed = False
        dynamic_title: typing.Optional[DynamicString.DynamicString] = None
        dynamic_title_d = self._get_persistent_property("dynamic_title").value
        if dynamic_title_d is not None:
            dynamic_title = DynamicString.make_dynamic_string(dynamic_title_d)
        if dynamic_title:
            self.__set_dynamic_title(dynamic_title)
        self.__pending_write = False

    def __set_dynamic_title(self, dynamic_title: typing.Optional[DynamicString.DynamicString]) -> None:
        self.__dynamic_title = dynamic_title
        if self.__dynamic_title:
            self.__dynamic_title.connect_item(self)
            self.__dynamic_title_persistence_action = Stream.ValueStreamAction(self.__dynamic_title.persistence_stream, ReferenceCounting.weak_partial(DataItem.__write_dynamic_title, self))
            self.__dynamic_title_stream.stream = self.__dynamic_title.string_stream
            self.__write_dynamic_title(self.__dynamic_title.persistence_stream.value)
        else:
            self.__dynamic_title_persistence_action = None
            self.__dynamic_title_stream.stream = None

    def __write_dynamic_title(self, dynamic_title_d: typing.Optional[Persistence.PersistentDictType]) -> None:
        if not self._is_reading:
            self._set_persistent_property_value("dynamic_title", dynamic_title_d)

    @property
    def dynamic_title(self) -> typing.Optional[DynamicString.DynamicString]:
        return self.__dynamic_title

    @dynamic_title.setter
    def dynamic_title(self, value: DynamicString.DynamicString) -> None:
        self.__set_dynamic_title(value)
        self._set_persistent_property_value("dynamic_title_enabled", True)

    def set_dynamic_title_by_id(self, dynamic_title_id: str) -> None:
        dynamic_string = DynamicString.make_dynamic_string({"type": dynamic_title_id})
        if dynamic_string:
            self.dynamic_title = dynamic_string

    @property
    def properties(self) -> typing.Optional[Persistence.PersistentDictType]:
        """ Used for debugging. """
        return self.get_storage_properties()

    @property
    def is_live(self) -> bool:
        """Return whether this library item represents live acquisition."""
        return self.__is_live

    def _enter_live_state(self) -> None:
        self.__is_live = True
        # live state is only a display item
        # also, there are still tests which filter data items; so leave this here until those tests are updated
        # the item_changed_event facilitates data item filtering
        self.item_changed_event.fire()

    def _exit_live_state(self) -> None:
        self.__is_live = False
        # live state is only a display item
        # also, there are still tests which filter data items; so leave this here until those tests are updated
        # the item_changed_event facilitates data item filtering
        self.item_changed_event.fire()

    def update_session(self, session_id: typing.Optional[str]) -> None:
        # update the session, but only if necessary (this is an optimization to prevent unnecessary display updates)
        if self.session_id != session_id:
            self.session_id = session_id
        session_metadata = ApplicationData.get_session_metadata_dict()
        if self.session_metadata != session_metadata:
            self.session_metadata = session_metadata

    class DataItemChangeContextManager:
        def __init__(self, data_item: DataItem) -> None:
            self.__data_item = data_item
        def __enter__(self) -> DataItem.DataItemChangeContextManager:
            self.__data_item._begin_data_item_changes()
            return self
        def __exit__(self, exception_type: typing.Optional[typing.Type[BaseException]], value: typing.Optional[BaseException], traceback: typing.Optional[types.TracebackType]) -> typing.Optional[bool]:
            self.__data_item._end_data_item_changes()
            return None

    def data_item_changes(self) -> contextlib.AbstractContextManager[DataItem.DataItemChangeContextManager]:
        # return a context manager to batch up a set of changes so that listeners
        # are only notified after the last change is complete.
        return DataItem.DataItemChangeContextManager(self)

    def _begin_data_item_changes(self) -> None:
        with self.__data_item_change_count_lock:
            change_count = self.__data_item_change_count
            self.__data_item_change_count += 1
        if change_count == 0:
            self.will_change_event.fire()

    def _end_data_item_changes(self) -> None:
        with self.__data_item_change_count_lock:
            self.__data_item_change_count -= 1
            change_count = self.__data_item_change_count
            content_changed = self.__content_changed
        # if the change count is now zero, it means that we're ready to notify listeners. but only notify listeners if
        # there are actual changes to report.
        if change_count == 0 and content_changed:
            self.item_changed_event.fire()
            self.did_change_event.fire()

    def __description_property_changed(self, name: str, value: typing.Any) -> None:
        self.__property_changed(name, value)
        self.notify_property_changed("date_for_sorting_local")
        self.__notify_description_changed()

    def __notify_description_changed(self) -> None:
        self._notify_data_item_content_changed()
        self._description_changed()

    def _description_changed(self) -> None:
        self.description_changed_event.fire()

    def __data_description_changed(self, name: str, value: int) -> None:
        self.__change_changed = True
        self.__property_changed(name, value)

    def __dimensional_calibrations_changed(self, name: str, value: typing.Any) -> None:
        self.__change_changed = True
        self.__property_changed(name, list(value.list))  # don't send out the CalibrationList object

    def __metadata_property_changed(self, name: str, value: typing.Any) -> None:
        self.__change_changed = True
        self.__property_changed(name, value)
        self.notify_property_changed("date_for_sorting_local")

    def __property_changed(self, name: str, value: typing.Any) -> None:
        if name == "title":
            self.__specified_title_stream.value = value
        if name == "dynamic_title_enabled":
            self.__dynamic_title_enabled_stream.value = value
        self.notify_property_changed(name)
        if name == "session":
            self.notify_property_changed("session_metadata")
        if name in ("title", "caption", "description"):
            self.__notify_description_changed()

    def __timezone_property_changed(self, name: str, value: typing.Optional[str]) -> None:
        timezone = self.timezone
        if self.__data_metadata and timezone is not None:
            self.__data_metadata._set_timezone(timezone)
            self.__data_metadata._set_timezone_offset(self.timezone_offset)
        self.notify_property_changed(name)

    # call this when the listeners need to be updated (via data_item_content_changed).
    # Calling this method will send the data_item_content_changed method to each listener by using the method
    # data_item_changes.
    def _notify_data_item_content_changed(self) -> None:
        with self.data_item_changes():
            with self.__data_item_change_count_lock:
                self.__content_changed = True

    # date times

    @property
    def date_for_sorting(self) -> datetime.datetime:
        data_modified = self.data_modified
        return data_modified if data_modified else self.created

    @property
    def date_for_sorting_local(self) -> datetime.datetime:
        date_utc = self.date_for_sorting
        tz_minutes = Utility.local_utcoffset_minutes(date_utc)
        return date_utc + datetime.timedelta(minutes=tz_minutes)

    @property
    def date_for_sorting_local_as_string(self) -> str:
        date_local = self.date_for_sorting_local
        return date_local.strftime("%c")

    @property
    def created_local_as_string(self) -> str:
        return self.created_local.strftime("%c")

    @property
    def created_local(self) -> datetime.datetime:
        created_utc = self.created
        tz_minutes = Utility.local_utcoffset_minutes(created_utc)
        return created_utc + datetime.timedelta(minutes=tz_minutes)

    # access description

    @property
    def _session_manager(self) -> typing.Optional[SessionManager]:
        return self.__session_manager

    def set_session_manager(self, session_manager: typing.Optional[SessionManager]) -> None:
        self.__session_manager = session_manager

    # override from storage to watch for changes to this library item. notify observers.
    def notify_property_changed(self, key: str) -> None:
        super().notify_property_changed(key)
        self._notify_data_item_content_changed()

    class ChangeContextManager:
        def __init__(self, begin_changes_fn: typing.Callable[[], None], end_changes_fn: typing.Callable[[], None]) -> None:
            self.__begin_changes_fn = begin_changes_fn
            self.__end_changes_fn = end_changes_fn
        def __enter__(self) -> DataItem.ChangeContextManager:
            self.__begin_changes_fn()
            return self
        def __exit__(self, exception_type: typing.Optional[typing.Type[BaseException]], value: typing.Optional[BaseException], traceback: typing.Optional[types.TracebackType]) -> typing.Optional[bool]:
            self.__end_changes_fn()
            return None

    def data_source_changes(self) -> contextlib.AbstractContextManager[DataItem.ChangeContextManager]:
        # return a context manager to batch up a set of changes so that listeners
        # are only notified after the last change is complete.
        return DataItem.ChangeContextManager(self.__begin_changes, self.__end_changes)

    def __begin_changes(self) -> None:
        self.will_change_event.fire()
        with self.__change_count_lock:
            if self.__change_count == 0:
                self.__change_thread = threading.current_thread()
            else:
                if self.__change_thread != threading.current_thread():
                    warnings.warn('begin changes from different threads', RuntimeWarning, stacklevel=2)
            self.__change_count += 1

    def __end_changes(self) -> None:
        changed = False
        data_changed = False
        with self.__change_count_lock:
            if self.__change_count == 1:
                self.__change_thread = None
            else:
                if self.__change_thread != threading.current_thread():
                    warnings.warn('end changes from different threads', RuntimeWarning, stacklevel=2)
            self.__change_count -= 1
            change_count = self.__change_count
            if change_count == 0:
                changed = self.__change_changed
                self.__change_changed = False
                data_changed = self.__change_data_changed
                self.__change_data_changed = False
        # if the change count is now zero, it means that we're ready
        # to pass on the next value.
        if change_count == 0:
            if data_changed:
                self.data_changed_event.fire()
            if not self._is_reading:
                if data_changed:
                    self.__write_delay_data_changed = True
                if data_changed or changed:
                    self._notify_data_item_content_changed()
                    self.data_item_changed_event.fire()
        self.did_change_event.fire()

    def increment_data_ref_count(self) -> int:
        with self.__data_ref_count_mutex:
            initial_count = self.__data_ref_count
            self.__data_ref_count += 1
            if not initial_count or self.__data is None:
                self.__load_data()
        return initial_count + 1

    def decrement_data_ref_count(self) -> int:
        with self.__data_ref_count_mutex:
            assert self.__data_ref_count > 0
            self.__data_ref_count -= 1
            final_count = self.__data_ref_count
            if not final_count:
                self.__unload_data()
        return final_count

    @property
    def _data_ref_count(self) -> int:
        return self.__data_ref_count

    def set_pending_xdata(self, xd: DataAndMetadata.DataAndMetadata) -> None:
        with self.__pending_xdata_lock:
            self.__pending_xdata = xd

    def queue_partial_update(self, partial_xdata: DataAndMetadata.DataAndMetadata, *, src_slice: typing.Sequence[slice],
                             dst_slice: typing.Sequence[slice], metadata: DataAndMetadata.DataMetadata) -> None:
        with self.__pending_xdata_lock:
            self.__pending_queue.append((partial_xdata, src_slice, dst_slice, metadata))

    def update_to_pending_xdata(self) -> None:
        with self.__pending_xdata_lock:
            pending_xdata = self.__pending_xdata
            pending_queue = self.__pending_queue
            self.__pending_xdata = None
            self.__pending_queue = list()
        if pending_xdata or pending_queue:
            assert threading.current_thread() == threading.main_thread()
            with self.data_item_changes():
                # it is an error to have both pending xdata and a pending queue
                assert not pending_xdata or not pending_queue
                if pending_xdata:
                    self.set_xdata(pending_xdata, data_modified=pending_xdata.timestamp)
                for partial_xdata, partial_src_slice, partial_dst_slice, partial_metadata in pending_queue:
                    # note: data_modified is only used in case that the data has not been updated yet. subsequent
                    # updates use the timestamp of the existing data.
                    self.set_data_and_metadata_partial(partial_metadata, partial_xdata, partial_src_slice, partial_dst_slice, update_metadata=True, data_modified=partial_metadata.timestamp)

    _data_count = 0

    @property
    def xdata(self) -> typing.Optional[DataAndMetadata.DataAndMetadata]:
        self.increment_data_ref_count()
        try:
            if self.__data_metadata and self.__data is not None:
                data_and_metadata = DataAndMetadata.DataAndMetadata(
                    data=self.__data,
                    data_shape_and_dtype=self.__data_metadata.data_shape_and_dtype,
                    intensity_calibration=self.__data_metadata.intensity_calibration,
                    dimensional_calibrations=self.__data_metadata.dimensional_calibrations,
                    metadata=self.__data_metadata.metadata,
                    timestamp=self.__data_metadata.timestamp,
                    data_descriptor=self.__data_metadata.data_descriptor,
                    timezone=self.__data_metadata.timezone,
                    timezone_offset=self.__data_metadata.timezone_offset
                )

                DataItem._data_count += 1

                def finalize() -> None:
                    DataItem._data_count -= 1

                weakref.finalize(data_and_metadata, finalize)

                return data_and_metadata

            return None
        finally:
            self.decrement_data_ref_count()

    @property
    def source_file_path(self) -> typing.Optional[pathlib.Path]:
        return self.__source_file_path

    @source_file_path.setter
    def source_file_path(self, value: typing.Optional[typing.Union[pathlib.Path, str]]) -> None:
        self.__source_file_path = pathlib.Path(value) if value is not None else pathlib.Path()

    @property
    def session_metadata(self) -> DataAndMetadata.MetadataType:
        return copy.deepcopy(self._get_persistent_property_value("session"))

    @session_metadata.setter
    def session_metadata(self, value: DataAndMetadata.MetadataType) -> None:
        self._set_persistent_property_value("session", copy.deepcopy(value))

    def _set_session_metadata(self, metadata: DataAndMetadata.MetadataType) -> None:
        self.session_metadata = metadata

    @property
    def session_data(self) -> DataAndMetadata.MetadataType:
        return self.session_metadata

    @session_data.setter
    def session_data(self, value: DataAndMetadata.MetadataType) -> None:
        self.session_metadata = value

    def __validate_session_id(self, value: typing.Any) -> typing.Optional[str]:
        if value is not None:
            assert isinstance(value, str)
            assert (value[8] == '-' and all(c.isdigit() for c in value[:8] + value[9:]))
            return value
        return value  # use two returns to satisfy mypy

    @property
    def data(self) -> typing.Optional[_ImageDataType]:
        return self.__get_data()

    def set_data(self, data: _ImageDataType, data_modified: typing.Optional[datetime.datetime] = None) -> None:
        timezone = Utility.get_local_timezone()
        timezone_offset = Utility.TimezoneMinutesToStringConverter().convert(Utility.local_utcoffset_minutes())
        self.set_xdata(DataAndMetadata.new_data_and_metadata(data, timestamp=data_modified, timezone=timezone, timezone_offset=timezone_offset))

    def set_xdata(self, xdata: typing.Optional[DataAndMetadata.DataAndMetadata], data_modified: typing.Optional[datetime.datetime] = None) -> None:
        if not self._closed:
            with self.data_source_changes():
                self.set_data_and_metadata(xdata, data_modified)

    class DataAccessor:
        def __init__(self, data_item: DataItem, get_data: typing.Callable[[], typing.Optional[_ImageDataType]], set_data: typing.Callable[[typing.Optional[_ImageDataType]], None]) -> None:
            self.__data_item = data_item
            self.__get_data = get_data
            self.__set_data = set_data

        def __enter__(self) -> DataItem.DataAccessor:
            self.__data_item.increment_data_ref_count()
            return self

        def __exit__(self, exception_type: typing.Optional[typing.Type[BaseException]], value: typing.Optional[BaseException], traceback: typing.Optional[types.TracebackType]) -> typing.Optional[bool]:
            self.__data_item.decrement_data_ref_count()
            return None

        @property
        def data(self) -> typing.Optional[_ImageDataType]:
            return self.__get_data()

        @data.setter
        def data(self, value: _ImageDataType) -> None:
            self.__set_data(value)

        def data_updated(self) -> None:
            self.__set_data(self.__get_data())

    # grab a data reference as a context manager. the object
    # returned defines data and data properties. reading data
    # should use the data property. writing data (if allowed) should
    # assign to the data property.
    def data_ref(self) -> contextlib.AbstractContextManager[DataItem.DataAccessor]:
        return DataItem.DataAccessor(self, self.__get_data, self.__set_data)

    def __get_data(self) -> typing.Optional[_ImageDataType]:
        xdata = self.xdata
        return xdata.data if xdata else None

    def __set_data(self, data: typing.Optional[_ImageDataType], data_modified: typing.Optional[datetime.datetime] = None) -> None:
        with self.data_source_changes():
            if data is not None:
                dimensional_shape = Image.dimensional_shape_from_data(data)
                data_and_metadata = self.data_and_metadata
                intensity_calibration = data_and_metadata.intensity_calibration if data_and_metadata else None
                dimensional_calibrations: typing.Optional[typing.List[Calibration.Calibration]] = None
                metadata: typing.Optional[DataAndMetadata.MetadataType] = None
                timestamp: typing.Optional[datetime.datetime] = None  # always update when the data is modified
                data_descriptor: typing.Optional[DataAndMetadata.DataDescriptor] = None
                if data_and_metadata:
                    assert dimensional_shape is not None
                    dimensional_calibrations = list(data_and_metadata.dimensional_calibrations)
                    while len(dimensional_calibrations) < len(dimensional_shape):
                        dimensional_calibrations.append(Calibration.Calibration())
                    while len(dimensional_calibrations) > len(dimensional_shape):
                        dimensional_calibrations.pop(-1)
                    metadata = data_and_metadata.metadata
                    data_descriptor = data_and_metadata.data_descriptor
                self.set_data_and_metadata(DataAndMetadata.new_data_and_metadata(data, intensity_calibration=intensity_calibration, dimensional_calibrations=dimensional_calibrations, metadata=metadata, timestamp=timestamp, data_descriptor=data_descriptor), data_modified)
            else:
                self.set_data_and_metadata(None)

    @property
    def intensity_calibration(self) -> typing.Optional[Calibration.Calibration]:
        return self.__data_metadata.intensity_calibration if self.__data_metadata else self.__intensity_calibration

    @intensity_calibration.setter
    def intensity_calibration(self, intensity_calibration: Calibration.Calibration) -> None:
        with self.data_source_changes():
            if self.__data_metadata:  # handle case of missing data and metadata but doing recording
                self.__data_metadata._set_intensity_calibration(intensity_calibration)
            self.__intensity_calibration = copy.deepcopy(intensity_calibration)  # backup in case of no data and metadata
            self._set_persistent_property_value("intensity_calibration", intensity_calibration)

    def set_intensity_calibration(self, intensity_calibration: Calibration.Calibration) -> None:
        self.intensity_calibration = intensity_calibration

    @property
    def dimensional_calibrations(self) -> DataAndMetadata.CalibrationListType:
        data_metadata = self.__data_metadata
        return copy.deepcopy(data_metadata.dimensional_calibrations) if data_metadata else self.__dimensional_calibrations

    @dimensional_calibrations.setter
    def dimensional_calibrations(self, dimensional_calibrations: DataAndMetadata.CalibrationListType) -> None:
        with self.data_source_changes():
            if self.__data_metadata:  # handle case of missing data and metadata but doing recording
                self.__data_metadata._set_dimensional_calibrations(dimensional_calibrations)
            self.__dimensional_calibrations = copy.deepcopy(list(dimensional_calibrations))  # backup in case of no data and metadata
            self._set_persistent_property_value("dimensional_calibrations", CalibrationList(dimensional_calibrations))

    def set_dimensional_calibrations(self, dimensional_calibrations: DataAndMetadata.CalibrationListType) -> None:
        self.dimensional_calibrations = dimensional_calibrations

    def set_dimensional_calibration(self, dimension: int, calibration: Calibration.Calibration) -> None:
        dimensional_calibrations = list(self.dimensional_calibrations)
        while len(dimensional_calibrations) <= dimension:
            dimensional_calibrations.append(Calibration.Calibration())
        dimensional_calibrations[dimension] = calibration
        self.set_dimensional_calibrations(dimensional_calibrations)

    @property
    def data_modified(self) -> typing.Optional[datetime.datetime]:
        return self.__data_metadata.timestamp if self.__data_metadata else None

    @data_modified.setter
    def data_modified(self, value: datetime.datetime) -> None:
        if self.__data_metadata:
            self.__data_metadata._set_timestamp(value)
            self._set_persistent_property_value("data_modified", value)
            self.__metadata_property_changed("data_modified", value)

    @property
    def timezone(self) -> str | None:
        data_metadata = self.__data_metadata
        return data_metadata.timezone if data_metadata else Utility.get_local_timezone()

    @timezone.setter
    def timezone(self, value: str | None) -> None:
        if self.__data_metadata:
            self.__data_metadata._set_timezone(value)
            self._set_persistent_property_value("timezone", value)
            self.__timezone_property_changed("timezone", value)

    @property
    def timezone_offset(self) -> str | None:
        data_metadata = self.__data_metadata
        return data_metadata.timezone_offset if data_metadata else Utility.TimezoneMinutesToStringConverter().convert(Utility.local_utcoffset_minutes())

    @timezone_offset.setter
    def timezone_offset(self, value: str | None) -> None:
        if self.__data_metadata:
            self.__data_metadata._set_timezone_offset(value)
            self._set_persistent_property_value("timezone_offset", value)
            self.__timezone_property_changed("timezone_offset", value)

    @property
    def metadata(self) -> DataAndMetadata.MetadataType:
        data_metadata = self.__data_metadata
        return dict(data_metadata.metadata) if data_metadata else copy.deepcopy(self.__metadata)

    @metadata.setter
    def metadata(self, metadata: DataAndMetadata.MetadataType) -> None:
        with self.data_source_changes():
            assert isinstance(metadata, dict)
            if self.__data_metadata:
                self.__data_metadata._set_metadata(metadata)
            self.__metadata = copy.deepcopy(metadata) if metadata else dict()
            self._set_persistent_property_value("metadata", self.__metadata)

    def _set_metadata(self, metadata: DataAndMetadata.MetadataType) -> None:
        self.metadata = metadata

    @property
    def has_data(self) -> bool:
        return self.__data_metadata is not None

    # used for testing
    @property
    def is_data_loaded(self) -> bool:
        return self.__data is not None

    @property
    def data_metadata(self) -> typing.Optional[DataAndMetadata.DataMetadata]:
        return self.__data_metadata

    @property
    def data_and_metadata(self) -> typing.Optional[DataAndMetadata.DataAndMetadata]:
        return self.xdata

    def __load_data(self) -> None:
        if self.persistent_object_context and self.__data is None and self.__data_metadata:
            self.__data = typing.cast(typing.Optional[_ImageDataType], self.read_external_data("data"))

    def __unload_data(self) -> None:
        if self.__data_and_metadata_unloadable:
            self.__data = None

    @property
    def is_unloadable(self) -> bool:
        return self.__data_and_metadata_unloadable

    def _force_unload(self) -> None:
        self.__data = None

    def __set_data_metadata_direct(self, data_metadata: DataAndMetadata.DataMetadata,
                                   data_modified: typing.Optional[datetime.datetime] = None) -> None:
        assert self.__data_ref_count > 0
        # set the data modified directly
        data_modified = data_modified if data_modified else DateTime.utcnow()
        data_metadata._set_timestamp(data_modified)
        # save the data_metadata. this must go before setting the persistent properties below because
        # of how the recorder works (grabs the attribute from data item).
        self.__data_metadata = copy.deepcopy(data_metadata)
        # set the persistent values for the data_metadata
        self._set_persistent_property_value("data_shape", data_metadata.data_shape)
        self._set_persistent_property_value("data_dtype", DtypeToStringConverter().convert(data_metadata.data_dtype))
        self._set_persistent_property_value("is_sequence", data_metadata.is_sequence)
        self._set_persistent_property_value("collection_dimension_count", data_metadata.collection_dimension_count)
        self._set_persistent_property_value("datum_dimension_count", data_metadata.datum_dimension_count)
        self._set_persistent_property_value("intensity_calibration", copy.deepcopy(data_metadata.intensity_calibration))
        self._set_persistent_property_value("dimensional_calibrations", CalibrationList(data_metadata.dimensional_calibrations))
        # save timezone info here so it doesn't get overwritten in intermediate states.
        timezone = data_metadata.timezone
        timezone_offset = data_metadata.timezone_offset
        if timezone:
            self._set_persistent_property_value("timezone", timezone)
        if timezone_offset:
            self._set_persistent_property_value("timezone_offset", timezone_offset)
        # explicitly set metadata into persistent storage to prevent notifications.
        self.__metadata = dict(data_metadata.metadata)
        self._set_persistent_property_value("metadata", self.__metadata)
        self._set_persistent_property_value("data_modified", data_modified)

    def __set_data_and_metadata_direct(self, data_and_metadata: typing.Optional[DataAndMetadata.DataAndMetadata],
                                       data_modified: typing.Optional[datetime.datetime] = None) -> None:
        assert self.__data_ref_count > 0
        self.__data = data_and_metadata.data if data_and_metadata else None
        if data_and_metadata:
            self.__set_data_metadata_direct(data_and_metadata.data_metadata, data_modified)
        self.__change_changed = True
        self.__change_data_changed = True
        if self._session_manager:
            session_id = self._session_manager.current_session_id
            self.session_id = session_id

    def set_data_and_metadata(self, data_and_metadata: typing.Optional[DataAndMetadata.DataAndMetadata],
                              data_modified: typing.Optional[datetime.datetime] = None) -> None:
        """Sets the underlying data and data-metadata to the data_and_metadata.

        Note: this does not make a copy of the data.
        """
        self.increment_data_ref_count()
        try:
            new_data_and_metadata: typing.Optional[DataAndMetadata.DataAndMetadata]
            if data_and_metadata:
                data = data_and_metadata.data
                data_shape_and_dtype = data_and_metadata.data_shape_and_dtype
                intensity_calibration = data_and_metadata.intensity_calibration
                dimensional_calibrations = data_and_metadata.dimensional_calibrations
                metadata = data_and_metadata.metadata
                timestamp = data_and_metadata.timestamp
                data_descriptor = data_and_metadata.data_descriptor
                timezone = data_and_metadata.timezone or Utility.get_local_timezone()
                timezone_offset = data_and_metadata.timezone_offset or Utility.TimezoneMinutesToStringConverter().convert(Utility.local_utcoffset_minutes())
                new_data_and_metadata = DataAndMetadata.DataAndMetadata(data=data, data_shape_and_dtype=data_shape_and_dtype, intensity_calibration=intensity_calibration, dimensional_calibrations=dimensional_calibrations, metadata=metadata, timestamp=timestamp, data_descriptor=data_descriptor, timezone=timezone, timezone_offset=timezone_offset)
            else:
                new_data_and_metadata = None
            self.__set_data_and_metadata_direct(new_data_and_metadata, data_modified)
            if self.__data is not None:
                if self.persistent_object_context and not self.is_write_delayed:
                    self.write_external_data("data", self.__data)
                    self.__data_and_metadata_unloadable = True
        finally:
            self.decrement_data_ref_count()

    def reserve_data(self, *, data_shape: DataAndMetadata.ShapeType, data_dtype: numpy.typing.DTypeLike, data_descriptor: DataAndMetadata.DataDescriptor, data_modified: typing.Optional[datetime.datetime] = None) -> None:
        """Reserves the underlying data without necessarily allocating memory. Useful for memory mapped files.
        """
        self.increment_data_ref_count()
        try:
            if self.persistent_object_context:
                self.reserve_external_data("data", data_shape, data_dtype)
                data_shape_and_dtype = data_shape, data_dtype
                timezone = Utility.get_local_timezone()
                timezone_offset = Utility.TimezoneMinutesToStringConverter().convert(Utility.local_utcoffset_minutes())
                data_metadata = DataAndMetadata.DataMetadata(data_shape_and_dtype=data_shape_and_dtype, data_descriptor=data_descriptor, metadata=self.metadata, timezone=timezone, timezone_offset=timezone_offset)
                self.__set_data_metadata_direct(data_metadata, data_modified)
                self.__load_data()
                self.__data_and_metadata_unloadable = True
                self.__data_and_metadata_first_update_after_reserve = True
        finally:
            self.decrement_data_ref_count()

    def set_data_and_metadata_partial(self, data_metadata: DataAndMetadata.DataMetadata,
                                      data_and_metadata: DataAndMetadata.DataAndMetadata, src: typing.Sequence[slice],
                                      dst: typing.Sequence[slice], update_metadata: bool = False,
                                      data_modified: typing.Optional[datetime.datetime] = None) -> None:
        # metadata is updated from data_metadata; data_and_metadata is only used for data
        with self.data_source_changes():
            self.increment_data_ref_count()
            try:
                if self.__data is None:
                    data: numpy.typing.NDArray[typing.Any] = numpy.zeros(data_metadata.data_shape, data_metadata.data_dtype)
                    data_shape_and_dtype = data_metadata.data_shape_and_dtype
                    intensity_calibration = data_metadata.intensity_calibration
                    dimensional_calibrations = data_metadata.dimensional_calibrations
                    metadata = data_metadata.metadata
                    timestamp = data_metadata.timestamp
                    data_descriptor = data_metadata.data_descriptor
                    timezone = data_metadata.timezone or Utility.get_local_timezone()
                    timezone_offset = data_metadata.timezone_offset or Utility.TimezoneMinutesToStringConverter().convert(
                        Utility.local_utcoffset_minutes())
                    new_data_and_metadata = DataAndMetadata.DataAndMetadata(data=data,
                                                                            data_shape_and_dtype=data_shape_and_dtype,
                                                                            intensity_calibration=intensity_calibration,
                                                                            dimensional_calibrations=dimensional_calibrations,
                                                                            metadata=metadata, timestamp=timestamp,
                                                                            data_descriptor=data_descriptor,
                                                                            timezone=timezone,
                                                                            timezone_offset=timezone_offset)
                    self.__set_data_and_metadata_direct(new_data_and_metadata, data_modified)
                if self.__data is not None:
                    date_modified = data_metadata.timestamp if self.__data_and_metadata_first_update_after_reserve else self.__data_metadata.timestamp if self.__data_metadata else None
                    self.__data_and_metadata_first_update_after_reserve = False
                    if update_metadata or date_modified:
                        self.__set_data_metadata_direct(data_metadata, date_modified)
                    assert self.data_shape == data_metadata.data_shape
                    assert self.data_dtype == data_metadata.data_dtype
                    assert self.data_dtype == data_and_metadata.data_dtype, f"{self.data_dtype=} == {data_and_metadata.data_dtype=}"
                    self.__data[tuple(dst)] = data_and_metadata._data_ex[tuple(src)]
                    # mark changes and update session
                    self.__change_changed = True
                    self.__change_data_changed = True
                    if self._session_manager:
                        session_id = self._session_manager.current_session_id
                        self.session_id = session_id
                    # set data_shape as a way to update 'modified' property
                    self._set_persistent_property_value("data_shape", self.data_shape)
                    if self.persistent_object_context and not self.is_write_delayed:
                        self.write_external_data("data", self.__data)
                        self.__data_and_metadata_unloadable = True
            finally:
                self.decrement_data_ref_count()

    @property
    def data_shape(self) -> typing.Optional[DataAndMetadata.ShapeType]:
        return self.__data_metadata.data_shape if self.__data_metadata else None

    @property
    def data_dtype(self) -> typing.Optional[numpy.typing.DTypeLike]:
        return self.__data_metadata.data_dtype if self.__data_metadata else None

    @property
    def data_descriptor(self) -> typing.Optional[DataAndMetadata.DataDescriptor]:
        return self.__data_metadata.data_descriptor if self.__data_metadata else None

    @property
    def dimensional_shape(self) -> DataAndMetadata.ShapeType:
        return self.__data_metadata.dimensional_shape if self.__data_metadata else tuple()

    @property
    def datum_dimension_shape(self) -> DataAndMetadata.ShapeType:
        return self.__data_metadata.datum_dimension_shape if self.__data_metadata else tuple()

    @property
    def datum_dimension_count(self) -> int:
        return self.__data_metadata.datum_dimension_count if self.__data_metadata else 0

    @property
    def collection_dimension_count(self) -> int:
        return self.__data_metadata.collection_dimension_count if self.__data_metadata else 0

    @property
    def is_collection(self) -> bool:
        return self.collection_dimension_count > 0 if self.collection_dimension_count is not None else False

    @property
    def is_sequence(self) -> bool:
        return self.__data_metadata.is_sequence if self.__data_metadata else False

    @property
    def is_data_1d(self) -> bool:
        return self.__data_metadata is not None and self.__data_metadata.is_data_1d

    @property
    def is_data_2d(self) -> bool:
        return self.__data_metadata is not None and self.__data_metadata.is_data_2d

    @property
    def is_data_3d(self) -> bool:
        return self.__data_metadata is not None and self.__data_metadata.is_data_3d

    @property
    def is_data_4d(self) -> bool:
        return self.__data_metadata is not None and self.__data_metadata.is_data_4d

    @property
    def is_data_rgb(self) -> bool:
        return self.__data_metadata is not None and self.__data_metadata.is_data_rgb

    @property
    def is_data_rgba(self) -> bool:
        return self.__data_metadata is not None and self.__data_metadata.is_data_rgba

    @property
    def is_datum_1d(self) -> bool:
        return self.__data_metadata is not None and self.__data_metadata.is_datum_1d

    @property
    def is_datum_2d(self) -> bool:
        return self.__data_metadata is not None and self.__data_metadata.is_datum_2d

    @property
    def is_data_rgb_type(self) -> bool:
        return self.__data_metadata is not None and self.__data_metadata.is_data_rgb_type

    @property
    def is_data_scalar_type(self) -> bool:
        return self.__data_metadata is not None and self.__data_metadata.is_data_scalar_type

    @property
    def is_data_complex_type(self) -> bool:
        return self.__data_metadata is not None and self.__data_metadata.is_data_complex_type

    @property
    def is_data_bool(self) -> bool:
        return self.__data_metadata is not None and self.__data_metadata.is_data_bool

    def get_data_value(self, pos: DataAndMetadata.ShapeType) -> typing.Any:
        xdata = self.xdata
        return xdata.get_data_value(pos) if xdata else None

    @property
    def size_and_data_format_as_string(self) -> str:
        return self.__data_metadata.size_and_data_format_as_string if self.__data_metadata else _("No Data")

    @property
    def storage_space_string(self) -> str:
        dimensional_shape = self.dimensional_shape
        data_dtype = self.data_dtype
        if dimensional_shape is not None and data_dtype is not None:

            # see https://stackoverflow.com/questions/12523586/python-format-size-application-converting-b-to-kb-mb-gb-tb
            def humanbytes(b: int) -> str:
                """Return the given bytes as a human friendly KB, MB, GB, or TB string."""
                B = float(b)
                KB = float(1024)
                MB = float(KB ** 2)  # 1,048,576
                GB = float(KB ** 3)  # 1,073,741,824
                TB = float(KB ** 4)  # 1,099,511,627,776

                if B < KB:
                    return '{0} {1}'.format(B, 'Bytes' if B != 1 else 'Byte')
                elif KB <= B < MB:
                    return '{0:.2f} KB'.format(B / KB)
                elif MB <= B < GB:
                    return '{0:.2f} MB'.format(B / MB)
                elif GB <= B < TB:
                    return '{0:.2f} GB'.format(B / GB)
                elif TB <= B:
                    return '{0:.2f} TB'.format(B / TB)
                else:
                    return "{0} Bytes".format(B)

            total_bytes = typing.cast(int, numpy.prod(dimensional_shape + (numpy.dtype(data_dtype).itemsize,), dtype=numpy.int64))

            return humanbytes(total_bytes)
        return str()

    @property
    def storage_string(self) -> str:
        storage_space_str = self.storage_space_string
        file_path = self.__get_file_path()
        path_suffix = pathlib.Path(file_path).suffix.lower() if file_path else None
        if path_suffix:
            return f"{storage_space_str} ({path_suffix[1:]})"
        else:
            return f"{storage_space_str}"

    def has_metadata_value(self, key: str) -> bool:
        return Metadata.has_metadata_value(self, key)

    def get_metadata_value(self, key: str) -> typing.Any:
        return Metadata.get_metadata_value(self, key)

    def set_metadata_value(self, key: str, value: typing.Any) -> None:
        Metadata.set_metadata_value(self, key, value)

    def delete_metadata_value(self, key: str) -> None:
        Metadata.delete_metadata_value(self, key)


def new_data_item(data_and_metadata_in: typing.Optional[DataAndMetadata._DataAndMetadataLike] = None) -> DataItem:
    data_and_metadata = DataAndMetadata.promote_ndarray(data_and_metadata_in) if data_and_metadata_in is not None else None
    data_item = DataItem(large_format=len(data_and_metadata.dimensional_shape) > 2 if data_and_metadata else False)
    data_item.set_xdata(data_and_metadata)
    return data_item

"""
Architectural Decision Records.

ADR 2024-04-27. The data item will separately keep track of whether dynamic title updating is enabled rather than
have it be dependent on the existence of the dynamic string object itself. This allows the user to enable or disable
the dynamic title without having to remember its previous state. This may be important for custom acquisition dynamic
strings or computed data item titling.
"""



from __future__ import annotations

import abc
import types
import typing

from nion.swift import Undo
from nion.swift.model import DisplayItem
from nion.swift.model import DocumentModel
from nion.swift.model import Graphics
from nion.swift.model import Persistence
from nion.swift.model import Utility
from nion.ui import CanvasItem
from nion.ui import DrawingContext
from nion.ui import UserInterface
from nion.utils import Geometry


class InteractiveTask:
    def __init__(self) -> None:
        pass

    def __enter__(self) -> InteractiveTask:
        return self

    def __exit__(self, exception_type: typing.Optional[typing.Type[BaseException]], value: typing.Optional[BaseException], traceback: typing.Optional[types.TracebackType]) -> typing.Optional[bool]:
        self.close()
        return None

    def close(self) -> None:
        self._close()

    def commit(self) -> None:
        self._commit()

    def _close(self) -> None: ...
    def _commit(self) -> None: ...


class DisplayCanvasItemDelegate(typing.Protocol):

    @property
    def tool_mode(self) -> str: raise NotImplementedError()

    @tool_mode.setter
    def tool_mode(self, value: str) -> None: ...

    def begin_mouse_tracking(self) -> None: ...
    def end_mouse_tracking(self, undo_command: typing.Optional[Undo.UndoableCommand]) -> None: ...
    def delete_key_pressed(self) -> None: ...
    def enter_key_pressed(self) -> None: ...
    def cursor_changed(self, pos: typing.Optional[typing.Tuple[int, ...]]) -> None: ...
    def update_display_properties(self, display_properties: Persistence.PersistentDictType) -> None: ...
    def update_display_data_channel_properties(self, display_data_channel_properties: Persistence.PersistentDictType) -> None: ...
    def create_change_display_command(self, *, command_id: typing.Optional[str] = None, is_mergeable: bool = False) -> Undo.UndoableCommand: ...
    def create_change_graphics_command(self) -> Undo.UndoableCommand: ...
    def create_insert_graphics_command(self, graphics: typing.Sequence[Graphics.Graphic], modified_state: typing.Any) -> Undo.UndoableCommand: ...
    def create_move_display_layer_command(self, display_item: DisplayItem.DisplayItem, src_index: int, target_index: int) -> Undo.UndoableCommand: ...
    def push_undo_command(self, command: Undo.UndoableCommand) -> None: ...
    def create_change_display_properties_task(self) -> InteractiveTask: ...
    def create_create_graphic_task(self, graphic_type: str, start_position: Geometry.FloatPoint) -> InteractiveTask: ...
    def create_change_graphics_task(self) -> InteractiveTask: ...
    def add_index_to_selection(self, index: int) -> None: ...
    def remove_index_from_selection(self, index: int) -> None: ...
    def set_selection(self, index: int) -> None: ...
    def clear_selection(self) -> None: ...
    def add_and_select_region(self, region: Graphics.Graphic) -> Undo.UndoableCommand: ...
    def nudge_selected_graphics(self, mapping: Graphics.CoordinateMappingLike, delta: Geometry.FloatSize) -> None: ...
    def nudge_slice(self, delta: int) -> None: ...
    def drag_graphics(self, graphics: typing.Sequence[Graphics.Graphic]) -> None: ...
    def adjust_graphics(self, widget_mapping: Graphics.CoordinateMappingLike, graphic_drag_items: typing.Sequence[Graphics.Graphic], graphic_drag_part: str, graphic_part_data: typing.Dict[int, Graphics.DragPartData], graphic_drag_start_pos: Geometry.FloatPoint, pos: Geometry.FloatPoint, modifiers: UserInterface.KeyboardModifiers) -> None: ...
    def display_clicked(self, modifiers: UserInterface.KeyboardModifiers) -> bool: ...
    def image_clicked(self, image_position: Geometry.FloatPoint, modifiers: UserInterface.KeyboardModifiers) -> bool: ...
    def image_mouse_pressed(self, image_position: Geometry.FloatPoint, modifiers: UserInterface.KeyboardModifiers) -> bool: ...
    def image_mouse_released(self, image_position: Geometry.FloatPoint, modifiers: UserInterface.KeyboardModifiers) -> bool: ...
    def image_mouse_position_changed(self, image_position: Geometry.FloatPoint, modifiers: UserInterface.KeyboardModifiers) -> bool: ...
    def show_display_context_menu(self, gx: int, gy: int) -> bool: ...
    def get_document_model(self) -> DocumentModel.DocumentModel: ...
    def create_rectangle(self, pos: Geometry.FloatPoint) -> Graphics.RectangleGraphic: ...
    def create_ellipse(self, pos: Geometry.FloatPoint) -> Graphics.EllipseGraphic: ...
    def create_line(self, pos: Geometry.FloatPoint) -> Graphics.LineGraphic: ...
    def create_point(self, pos: Geometry.FloatPoint) -> Graphics.PointGraphic: ...
    def create_line_profile(self, pos: Geometry.FloatPoint) -> Graphics.LineProfileGraphic: ...
    def create_spot(self, pos: Geometry.FloatPoint) -> Graphics.SpotGraphic: ...
    def create_wedge(self, angle: float) -> Graphics.WedgeGraphic: ...
    def create_ring(self, radius: float) -> Graphics.RingGraphic: ...
    def create_lattice(self, u_pos: Geometry.FloatSize) -> Graphics.LatticeGraphic: ...


class DisplayCanvasItem(CanvasItem.CanvasItemComposition):
    """Allow display canvas items (tools) to bypass multi-select.

    In cases where only_shift is used in the tool, the tool should bypass multi-select and the result will be as if
    the user clicked in the target display panel without modifiers. For instance, if the user shift-clicks using the
    crop tool to restrict the crop to a square shape, the crop tool should bypass multi-select and the result will be
    as if the user clicked in the target display panel without modifiers, selecting the target panel.
    """
    bypass_multi_select = False

    @property
    def key_contexts(self) -> typing.Sequence[str]:
        """Return key contexts.

        Key contexts provide an ordered list of contexts that are used to determine
        which actions are valid at a given time. The contexts are checked in reverse
        order (i.e. last added have highest precedence).
        """
        return list()

    @property
    def mouse_mapping(self) -> Graphics.CoordinateMappingLike: raise NotImplementedError()

    @abc.abstractmethod
    def add_display_control(self, display_control_canvas_item: CanvasItem.AbstractCanvasItem, role: typing.Optional[str] = None) -> None: ...

    @abc.abstractmethod
    def handle_auto_display(self) -> bool: ...

    def update_display_data_delta(self, display_data_delta: DisplayItem.DisplayDataDelta) -> None: ...

    def set_focused(self, is_focused: bool) -> None: ...

    def _wait_for_update(self) -> None:
        pass


class FrameRateCanvasItemComposer(CanvasItem.BaseComposer):
    def __init__(self, canvas_item: CanvasItem.AbstractCanvasItem, layout_sizing: CanvasItem.Sizing, cache: CanvasItem.ComposerCache, fps: str, fps2: str, fps3: str) -> None:
        super().__init__(canvas_item, layout_sizing, cache)
        self.__fps = fps
        self.__fps2 = fps2
        self.__fps3 = fps3

    def _repaint(self, drawing_context: DrawingContext.DrawingContext, canvas_bounds: Geometry.IntRect, composer_cache: CanvasItem.ComposerCache) -> None:
        fps = self.__fps
        fps2 = self.__fps2
        fps3 = self.__fps3
        with drawing_context.saver():
            drawing_context.translate(canvas_bounds.left, canvas_bounds.top)
            font = "normal 11px serif"
            text_pos = canvas_bounds.top_left
            drawing_context.begin_path()
            drawing_context.move_to(text_pos.x, text_pos.y)
            drawing_context.line_to(text_pos.x + 200, text_pos.y)
            drawing_context.line_to(text_pos.x + 200, text_pos.y + 60)
            drawing_context.line_to(text_pos.x, text_pos.y + 60)
            drawing_context.close_path()
            drawing_context.fill_style = "rgba(255, 255, 255, 0.6)"
            drawing_context.fill()
            drawing_context.font = font
            drawing_context.text_baseline = "middle"
            drawing_context.text_align = "left"
            drawing_context.fill_style = "#000"
            drawing_context.fill_text("display:" + fps, text_pos.x + 8, text_pos.y + 10)
            drawing_context.fill_text("frame:" + fps2, text_pos.x + 8, text_pos.y + 30)
            drawing_context.fill_text("update:" + fps3, text_pos.x + 8, text_pos.y + 50)
            drawing_context.statistics("display")


class FrameRateCanvasItem(CanvasItem.AbstractCanvasItem):
    """Display frame rate in the display canvas item.

    There are three rates that are tracked. The "display" rate is the rate at which the display is updated and is
    "ticked" when low level repaint occurs, which generally happens once per update. The "frame" rate is the rate at
    the frame index changes. The "update" rate is the rate at which the display is updated and is "ticked" when the
    display is updated. It is different from the "display" since it is just a request to update; whereas the "display"
    is the actual repaint being performed.
    """
    def __init__(self) -> None:
        super().__init__()
        self.__display_frame_rate_id: typing.Optional[str] = None
        self.__display_frame_rate_last_index = 0

    @property
    def display_frame_rate_id(self) -> typing.Optional[str]:
        return self.__display_frame_rate_id

    @display_frame_rate_id.setter
    def display_frame_rate_id(self, value: typing.Optional[str]) -> None:
        self.__display_frame_rate_id = value
        self.update()

    def toggle_display(self, display_frame_rate_id: str) -> None:
        if self.__display_frame_rate_id is None:
            self.__display_frame_rate_id = display_frame_rate_id
        else:
            self.__display_frame_rate_id = None
        self.update()

    def frame_tick(self, frame_index: int) -> None:
        if self.__display_frame_rate_id:
            if frame_index != self.__display_frame_rate_last_index:
                Utility.fps_tick("frame_" + self.__display_frame_rate_id)
                self.__display_frame_rate_last_index = frame_index
            Utility.fps_tick("update_" + self.__display_frame_rate_id)
            self.update()

    def _get_composer(self, composer_cache: CanvasItem.ComposerCache) -> typing.Optional[CanvasItem.BaseComposer]:
        if self.__display_frame_rate_id:
            fps = Utility.fps_get("display_" + self.__display_frame_rate_id)
            fps2 = Utility.fps_get("frame_" + self.__display_frame_rate_id)
            fps3 = Utility.fps_get("update_" + self.__display_frame_rate_id)
            return FrameRateCanvasItemComposer(self, self.sizing, composer_cache, fps, fps2, fps3)
        return CanvasItem.EmptyCanvasItemComposer(self, self.sizing, composer_cache)




# standard libraries
import contextlib
import copy
import datetime
import functools
import gc
import logging
import math
import threading
import time
import typing
import unittest
import weakref

# third party libraries
import numpy

# local libraries
from nion.data import Calibration
from nion.data import DataAndMetadata
from nion.swift import Application
from nion.swift.model import DataItem
from nion.swift.model import Metadata
from nion.swift.test import TestContext
from nion.ui import TestUI


def create_memory_profile_context() -> TestContext.MemoryProfileContext:
    return TestContext.MemoryProfileContext()


class TestDataItemClass(unittest.TestCase):

    def setUp(self):
        TestContext.begin_leaks()
        self.app = Application.Application(TestUI.UserInterface(), set_global=False)

    def tearDown(self):
        TestContext.end_leaks(self)

    def test_metadata_value_data_item_methods(self) -> None:
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((4, 4), float))
            document_model.append_data_item(data_item)
            # test a metadata value
            self.assertFalse(data_item.has_metadata_value("stem.high_tension"))
            data_item.set_metadata_value("stem.high_tension", 100000.0)
            self.assertEqual(100000.0, data_item.metadata["instrument"]["high_tension"])
            self.assertTrue(data_item.has_metadata_value("stem.high_tension"))
            self.assertEqual(100000.0, data_item.get_metadata_value("stem.high_tension"))
            data_item.delete_metadata_value("stem.high_tension")
            self.assertFalse(data_item.has_metadata_value("stem.high_tension"))
            # test a session metadata value
            self.assertFalse(data_item.has_metadata_value("stem.session.sample"))
            data_item.set_metadata_value("stem.session.sample", "Unobtainium")
            self.assertEqual("Unobtainium", data_item.session_metadata["sample"])
            self.assertTrue(data_item.has_metadata_value("stem.session.sample"))
            self.assertEqual("Unobtainium", data_item.get_metadata_value("stem.session.sample"))
            data_item.delete_metadata_value("stem.session.sample")
            self.assertFalse(data_item.has_metadata_value("stem.session.sample"))

    def test_metadata_value_xdata_methods(self) -> None:
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((4, 4), float))
            document_model.append_data_item(data_item)
            xdata = data_item.xdata
            self.assertFalse(Metadata.has_metadata_value(xdata, "stem.high_tension"))
            Metadata.set_metadata_value(xdata, "stem.high_tension", 100000.0)
            self.assertEqual(100000.0, xdata.metadata["instrument"]["high_tension"])
            self.assertTrue(Metadata.has_metadata_value(xdata, "stem.high_tension"))
            self.assertEqual(100000.0, Metadata.get_metadata_value(xdata, "stem.high_tension"))
            Metadata.delete_metadata_value(xdata, "stem.high_tension")
            self.assertFalse(Metadata.has_metadata_value(xdata, "stem.high_tension"))
            # test a session metadata value
            self.assertFalse(Metadata.has_metadata_value(xdata, "stem.session.sample"))
            Metadata.set_metadata_value(xdata, "stem.session.sample", "Unobtainium")
            self.assertEqual("Unobtainium", xdata.metadata["session_metadata"]["sample"])
            self.assertTrue(Metadata.has_metadata_value(xdata, "stem.session.sample"))
            self.assertEqual("Unobtainium", Metadata.get_metadata_value(xdata, "stem.session.sample"))
            Metadata.delete_metadata_value(xdata, "stem.session.sample")
            self.assertFalse(Metadata.has_metadata_value(xdata, "stem.session.sample"))

    def test_metadata_value_dict_methods(self) -> None:
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((4, 4), float))
            document_model.append_data_item(data_item)
            d = dict()
            self.assertFalse(Metadata.has_metadata_value(d, "stem.high_tension"))
            Metadata.set_metadata_value(d, "stem.high_tension", 100000.0)
            self.assertEqual(100000.0, d["instrument"]["high_tension"])
            self.assertTrue(Metadata.has_metadata_value(d, "stem.high_tension"))
            self.assertEqual(100000.0, Metadata.get_metadata_value(d, "stem.high_tension"))
            Metadata.delete_metadata_value(d, "stem.high_tension")
            self.assertFalse(Metadata.has_metadata_value(d, "stem.high_tension"))
            # test a session metadata value
            self.assertFalse(Metadata.has_metadata_value(d, "stem.session.sample"))
            Metadata.set_metadata_value(d, "stem.session.sample", "Unobtainium")
            self.assertEqual("Unobtainium", d["session_metadata"]["sample"])
            self.assertTrue(Metadata.has_metadata_value(d, "stem.session.sample"))
            self.assertEqual("Unobtainium", Metadata.get_metadata_value(d, "stem.session.sample"))
            Metadata.delete_metadata_value(d, "stem.session.sample")
            self.assertFalse(Metadata.has_metadata_value(d, "stem.session.sample"))


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.DEBUG)
    unittest.main()



from __future__ import annotations

# standard libraries
import asyncio
import dataclasses
import datetime
import functools
import gettext
import json
import logging
import operator
import os
import pathlib
import pkgutil
import re
import sys
import typing
import urllib.parse
import uuid
import webbrowser

# third party libraries
# None

# local libraries
from nion.swift import ActivityPanel
from nion.swift import DataPanel
from nion.swift import DocumentController
from nion.swift import FilterPanel
from nion.swift import HistogramPanel
from nion.swift import InfoPanel
from nion.swift import Inspector
from nion.swift import MetadataPanel
from nion.swift import NotificationDialog
from nion.swift import Panel
from nion.swift import ProjectPanel
from nion.swift import SessionPanel
from nion.swift import Task
from nion.swift import ToolbarPanel
from nion.swift import Workspace
from nion.swift.model import ApplicationData
from nion.swift.model import ColorMaps
from nion.swift.model import DocumentModel
from nion.swift.model import Feature
from nion.swift.model import FileStorageSystem
from nion.swift.model import PlugInManager
from nion.swift.model import Profile
from nion.swift.model import Symbolic
from nion.swift.model import Utility
from nion.ui import Application as UIApplication
from nion.ui import CanvasItem
from nion.ui import Declarative
from nion.ui import Dialog
from nion.ui import PreferencesDialog
from nion.ui import UserInterface
from nion.ui import Window as UIWindow
from nion.utils import Event
from nion.utils import Geometry
from nion.utils import ListModel
from nion.utils import Model
from nion.utils import Registry
from nion.utils import Stream

if typing.TYPE_CHECKING:
    from nion.swift.model import DisplayItem

_ = gettext.gettext

app: Application = typing.cast("Application", None)


class PersistenceHandler(UserInterface.PersistenceHandler):
    # handle persistent values using a file instead of ui-tool.

    def get_string(self, key: str) -> typing.Tuple[bool, str]:
        application_data = ApplicationData.get_data()
        values = application_data.get("preference-values", dict())
        if key in values:
            return True, values.get(key)
        return False, str()

    def set_string(self, key: str, value: str) -> bool:
        application_data = ApplicationData.get_data()
        values = application_data.setdefault("preference-values", dict())
        values[key] = value
        ApplicationData.set_data(application_data)
        return True

    def remove_key(self, key: str) -> bool:
        application_data = ApplicationData.get_data()
        values = application_data.get("preference-values", dict())
        if key in values:
            values.pop(key)
            ApplicationData.set_data(application_data)
            return True
        return False


class AboutDialog(Dialog.OkCancelDialog):
    def __init__(self, ui: UserInterface.UserInterface, parent_window: UIWindow.Window, version_str: str):
        super().__init__(ui, include_cancel=False, parent_window=parent_window)

        class Handler(Declarative.Handler):
            def __init__(self) -> None:
                super().__init__()

                logo_png = pkgutil.get_data(__name__, "resources/Logo3.png")
                assert logo_png is not None

                self.icon = CanvasItem.load_rgba_data_from_bytes(logo_png)

                changes_json_data = pkgutil.get_data(__name__, "resources/changes.json")
                assert changes_json_data is not None
                changes_data: typing.Sequence[typing.Mapping[str, typing.Any]] = json.loads(changes_json_data)

                u = Declarative.DeclarativeUI()

                base_prefix_label = [u.create_spacing(13), u.create_label(text=sys.base_prefix)] if sys.base_prefix else [u.create_spacing(0)]

                package_names = []
                package_versions = []

                for __, package_name, package_version in Application.get_nion_swift_version_info():
                    package_names.append(u.create_label(text=package_name))
                    package_versions.append(u.create_label(text=package_version))

                self.markdown = str()

                for version_d in changes_data:
                    changes_version_str = version_d["version"]
                    release_date_str = version_d.get("release_date")
                    self.markdown += "\n" + "### " + changes_version_str + (f" ({release_date_str})" if release_date_str else "") + "\n"
                    for notes_d in version_d.get("notes", list()):
                        summary_line = notes_d["summary"]
                        issue_thunks = list()
                        for issue_url in notes_d.get("issues", list()):
                            numbers = re.findall(r"\d+", issue_url)
                            if numbers:
                                issue_number = numbers[-1]
                                issue_thunks.append(f"[#{issue_number}]({issue_url})")
                        if issue_thunks:
                            summary_line += " (" + ", ".join(issue_thunks) + ")"
                        author_thunks = list()
                        for author_d in notes_d.get("authors", list()):
                            author = author_d.get("github")
                            if author:
                                author_thunks.append(f"[{author}](https://github.com/{author})")
                        if author_thunks:
                            summary_line += " by " + ", ".join(author_thunks)
                        self.markdown += "* " + summary_line + "\n"

                packages_scroll_area = u.create_scroll_area(u.create_row(u.create_column(*package_names), u.create_spacing(8),
                                                         u.create_column(*package_versions), u.create_stretch(),
                                                         spacing=8))
                icon_column = u.create_column(u.create_image(image="icon", width=72, height=72), u.create_stretch(), margin=4, spacing=8)
                main_content_column = u.create_column(u.create_label(text=f"Nion Swift {version_str}"),
                                                      u.create_label(text="Copyright Bruker 2025. All Rights Reserved."),
                                                      u.create_spacing(13),
                                                      u.create_label(text=f"Python {sys.version}", word_wrap=True),
                                                      *base_prefix_label, packages_scroll_area, u.create_stretch(),
                                                      margin_top=4,
                                                      margin_bottom=4,
                                                      spacing=8)
                recent_changes_content_column = u.create_column(u.create_scroll_area(u.create_column(
                    u.create_text_browser(markdown="@binding(markdown)", on_anchor_clicked="handle_anchor_clicked", size_policy_vertical="expanding"))))
                main_tabs = u.create_tabs(
                    u.create_tab(label=_("About"), content=main_content_column),
                    u.create_tab(label=_("Recent Changes"), content=recent_changes_content_column),
                    style="minimal"
                )
                self.ui_view = u.create_column(
                    u.create_row(
                        icon_column,
                        main_tabs,
                        spacing=8,
                        margin=6
                    ),
                    min_width=640,
                    min_height=300
                )

            def handle_anchor_clicked(self, widget: Declarative.UIWidget, anchor: str) -> bool:
                o = urllib.parse.urlparse(anchor)
                if o.scheme in ("http", "https"):
                    webbrowser.open(anchor)
                    return True
                return False

        self.content.add(Declarative.DeclarativeWidget(ui, self.event_loop or asyncio.get_event_loop(), Handler()))


# facilitate bootstrapping the application
class Application(UIApplication.BaseApplication):
    count = 0  # useful for detecting leaks in tests

    def __init__(self, ui: UserInterface.UserInterface, set_global: bool = True) -> None:
        super().__init__(ui)
        self.__class__.count += 1

        # reset these values for tests. otherwise tests run slower after app.start is called in any previous test.
        Symbolic.computation_min_period = 0.0
        Symbolic.computation_min_factor = 0.0

        logging.getLogger("migration").setLevel(logging.ERROR)
        logging.getLogger("loader").setLevel(logging.ERROR)

        logging.getLogger("_commands").disabled = True

        ui.set_application_info("Nion Swift", "Nion", "nion.com")

        ui.set_persistence_handler(PersistenceHandler())
        setattr(self.ui, "persistence_root", "3")  # sets of preferences
        self.version_str = "16.16.0"

        self.document_model_available_event = Event.Event()

        global app
        app = self  # hack to get the single instance set. hmm. better way?

        self.__profile: typing.Optional[Profile.Profile] = None
        self.__document_model: typing.Optional[DocumentModel.DocumentModel] = None

        self.__menu_handlers: typing.List[typing.Callable[[DocumentController.DocumentController], None]] = []

        Registry.register_component(Inspector.DeclarativeImageChooserConstructor(self), {"declarative_constructor"})
        Registry.register_component(Inspector.DeclarativeDataSourceChooserConstructor(self), {"declarative_constructor"})
        Registry.register_component(Inspector.DeclarativeTextPushButtonWidgetConstructor(self), {"declarative_constructor"})
        Registry.register_component(Inspector.DeclarativeColorChooserConstructor(self), {"declarative_constructor"})

        workspace_manager = Workspace.WorkspaceManager()
        workspace_manager.register_panel(SessionPanel.SessionPanel, "session-panel", _("Session"), ["left", "right"], "right", {"min-width": 320, "height": 80})
        workspace_manager.register_panel(ProjectPanel.CollectionsPanel, "collections-panel", _("Collections"), ["left", "right"], "left", {"min-width": 320, "min-height": 200})
        workspace_manager.register_panel(DataPanel.DataPanel, "data-panel", _("Data Panel"), ["left", "right"], "left", {"min-width": 320, "min-height": 320})
        workspace_manager.register_panel(HistogramPanel.HistogramPanel, "histogram-panel", _("Histogram"), ["left", "right"], "right", {"min-width": 320, "height": 140})
        workspace_manager.register_panel(InfoPanel.InfoPanel, "info-panel", _("Info"), ["left", "right"], "right", {"min-width": 320, "height": 60})
        workspace_manager.register_panel(Inspector.InspectorPanel, "inspector-panel", _("Inspector"), ["left", "right"], "right", {"min-width": 320})
        workspace_manager.register_panel(Task.TaskPanel, "task-panel", _("Task Panel"), ["left", "right"], "right", {"min-width": 320})
        workspace_manager.register_panel(Panel.OutputPanel, "output-panel", _("Output"), ["bottom"], "bottom", {"min-width": 480, "min-height": 200})
        workspace_manager.register_panel(ToolbarPanel.ToolbarPanel, "toolbar-panel", _("Toolbar"), ["top"], "top", {"height": 30})
        workspace_manager.register_panel(MetadataPanel.MetadataPanel, "metadata-panel", _("Metadata"), ["left", "right"], "right", {"width": 320, "height": 8})
        workspace_manager.register_panel(ActivityPanel.ActivityPanel, "activity-panel", _("Activity"), ["left", "right"], "right", {"min-width": 320, "height": 80})
        workspace_manager.register_filter_panel(FilterPanel.FilterPanel)

    def __purge_log_files(self, log_dir_path: pathlib.Path) -> None:
        try:
            @dataclasses.dataclass
            class LogFile:
                absolute_file_path: pathlib.Path
                time_delta: datetime.timedelta

            if log_dir_path.exists():
                log_files = list[LogFile]()
                now = datetime.datetime.now()
                for file_path in log_dir_path.rglob("commands*.log*"):
                    log_files.append(LogFile(file_path, now - datetime.datetime.fromtimestamp(file_path.stat().st_mtime)))
                log_files.sort(key=operator.attrgetter("time_delta"))
                for index, log_file in enumerate(log_files):
                    if index >= 100 or log_file.time_delta.days > 30:  # note: index is number from 0
                        logging.info(f"Purging log file {log_file.absolute_file_path}")
                        log_file.absolute_file_path.unlink()
        except Exception as e:
            logging.error(f"Error purging log files: {e}")
            # do not raise here; otherwise an error here will prevent launch.

    def initialize(self, *, load_plug_ins: bool = True, use_root_dir: bool = True) -> None:
        super().initialize()
        NotificationDialog._app = self
        # configure app data
        if load_plug_ins:
            now_slug = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
            log_dir_path = pathlib.Path(self.ui.get_data_location()) / "Logs"
            log_dir_path.mkdir(parents=True, exist_ok=True)
            log_file_path = log_dir_path / f"commands_{now_slug}.log"
            log_formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(threadName)s - %(message)s')
            log_file_handler = logging.FileHandler(log_file_path)
            log_file_handler.setFormatter(log_formatter)
            commands_logger = logging.getLogger("_commands")
            commands_logger.propagate = False
            commands_logger.disabled = False
            commands_logger.setLevel(logging.DEBUG)
            for handler in list(commands_logger.handlers):
                commands_logger.removeHandler(handler)
            commands_logger.addHandler(log_file_handler)

            python_version = str(sys.version.replace('\n', ''))

            COMMANDS_LOGGER_VERSION: typing.Final[str] = "1"
            commands_logger.info(f"# application launched (version {COMMANDS_LOGGER_VERSION})")
            commands_logger.info(f"# python version {python_version}")
            commands_logger.info(f"# ui {type(self.ui).__name__} / {type(getattr(self.ui, 'proxy')).__name__} / {self.ui.get_build_version()}")
            commands_logger.info(f"# qt {self.ui.get_qt_version()}")

            logging.info(f"Launch time {datetime.datetime.now()}")
            logging.info(f"Python version {python_version}")
            logging.info(f"User interface class {type(self.ui).__name__} / {type(getattr(self.ui, 'proxy')).__name__}")
            logging.info(f"Build version (UI) {self.ui.get_build_version()}")
            logging.info(f"Qt version {self.ui.get_qt_version()}")
            self.__purge_log_files(log_dir_path)
            logging.info(f"Log file {log_file_path} (level {logging.getLevelName(commands_logger.level)})")
            app_data_file_path = self.ui.get_configuration_location() / pathlib.Path("nionswift_appdata.json")
            ApplicationData.set_file_path(app_data_file_path)
            logging.info("Application data: " + str(app_data_file_path))
            self.__initialize_features()
            PlugInManager.load_plug_ins(self.ui.get_document_location(), self.ui.get_data_location(), get_root_dir() if use_root_dir else None)
            color_maps_dir = self.ui.get_configuration_location() / pathlib.Path("Color Maps")
            if color_maps_dir.exists():
                logging.info("Loading color maps from " + str(color_maps_dir))
                ColorMaps.load_color_maps(color_maps_dir)
            else:
                logging.info("NOT Loading color maps from " + str(color_maps_dir) + " (missing)")
        Registry.register_component(self, {"application"})

    def deinitialize(self) -> None:
        # shut down hardware source manager, unload plug-ins, and really exit ui
        NotificationDialog.close_notification_dialog()
        Registry.unregister_component(self, {"application"})
        if self.__profile:
            self.__profile.close()
            self.__profile = None
        self.__document_model = None
        PlugInManager.unload_plug_ins()
        commands_logger = logging.getLogger("_commands")
        commands_logger.info("# application shutdown")
        global app
        app = typing.cast(Application, None)  # hack to get the single instance set. hmm. better way?
        self.__class__.count -= 1
        super().deinitialize()

    def __initialize_features(self) -> None:
        enabled_feature_str = UserInterface.StringPersistentModel(self.ui, "enabled_feature_str", "")

        feature_manager = Feature.FeatureManager()
        feature_manager.enabled_feature_str = enabled_feature_str.value or str()

        def handle_feature_manager_property_changed(property_name: str) -> None:
            if property_name == "enabled_feature_str":
                enabled_feature_str.value = feature_manager.enabled_feature_str

        self.__enabled_feature_str_listener = feature_manager.property_changed_event.listen(handle_feature_manager_property_changed)

    @property
    def profile(self) -> Profile.Profile:
        assert self.__profile
        return self.__profile

    @property
    def document_model(self) -> DocumentModel.DocumentModel:
        assert self.__document_model
        return self.__document_model

    # for testing
    def _set_document_model(self, document_model: DocumentModel.DocumentModel) -> None:
        self.__document_model = document_model

    def start(self, *, profile_dir: typing.Optional[pathlib.Path] = None, profile: typing.Optional[Profile.Profile] = None) -> bool:
        """Start the application.

        Creates the profile object using profile_path parameter (for testing), the profile path constructed from the
        name stored in preferences, or a default profile path.

        Attaches recent projects if profile is freshly created. The recent projects will initially be disabled and
        will require the user to explicitly upgrade them.

        Creates the document model with the profile.

        Creates the document window (aka document controller) with the document model.
        """
        logging.getLogger("migration").setLevel(logging.INFO)
        logging.getLogger("loader").setLevel(logging.INFO)

        # create or load the profile object. allow test to override profile.
        is_created = False
        if not profile:
            # determine the profile_path
            if profile_dir:
                profile_path = profile_dir / pathlib.Path("Profile").with_suffix(".nsproj")
            else:
                data_dir = pathlib.Path(self.ui.get_data_location())
                profile_name = pathlib.Path(self.ui.get_persistent_string("profile_name", "Profile"))
                profile_path = data_dir / profile_name.with_suffix(".nsproj")
            # create the profile
            profile, is_created = self.__establish_profile(profile_path)
        self.__profile = profile
        assert self.__profile

        # test code to reset script items
        # self.__profile.script_items_updated = False
        # while self.__profile.script_items:
        #     self.__profile.remove_script_item(self.__profile.script_items[-1])

        # migrate the interactive scripts persistent object
        if not self.__profile.script_items_updated:
            items_str = self.ui.get_persistent_string("interactive_scripts_1")
            if items_str:
                for item_dict in json.loads(items_str):
                    item_type = item_dict.get("__type__", None)
                    if item_type == "FolderListItem":
                        folder_path_str = item_dict.get("full_path", None)
                        folder_path = pathlib.Path(folder_path_str) if folder_path_str else None
                        if folder_path:
                            self.__profile.append_script_item(Profile.FolderScriptItem(pathlib.Path(folder_path)))
                    elif item_type == "ScriptListItem" and item_dict.get("indent_level", None) == 0:
                        script_path_str = item_dict.get("full_path", None)
                        script_path = pathlib.Path(script_path_str) if script_path_str else None
                        if script_path:
                            self.__profile.append_script_item(Profile.FileScriptItem(pathlib.Path(script_path)))
            else:
                items_old = self.ui.get_persistent_object("interactive_scripts_0", list())
                for script_path_str in items_old:
                    script_path = pathlib.Path(script_path_str) if script_path_str else None
                    if script_path:
                        self.__profile.append_script_item(Profile.FileScriptItem(pathlib.Path(script_path)))
            self.__profile.script_items_updated = True

        # configure the document model object.
        Symbolic.computation_min_period = 0.1
        Symbolic.computation_min_factor = 1.0

        # if it was created, it probably means it is migrating from an old version. so add all recent projects.
        # they will initially be disabled and the user will have to explicitly upgrade them.
        if is_created:
            # present a dialog for showing progress while finding existing projects
            u = Declarative.DeclarativeUI()
            task_message = u.create_label(text=_("Looking for existing projects..."))
            progress = u.create_progress_bar(value="@binding(progress_value_model.value)", width=300 - 24)
            progress_message = u.create_label(text="@binding(message_str_model.value)")
            main_column = u.create_column(task_message, progress, progress_message, spacing=8, width=300)
            window = u.create_window(main_column, title=_("Locating Existing Projects"), margin=12, window_style="tool")

            # handler for the progress window. defines two value models: progress, an int 0-100; and message, a string.
            class FindExistingProjectsWindowHandler(Declarative.WindowHandler):
                def __init__(self, *, completion_fn: typing.Optional[typing.Callable[[], None]] = None):
                    super().__init__(completion_fn=completion_fn)
                    self.progress_value_model = Model.PropertyModel(0)
                    self.message_str_model = Model.PropertyModel(str())

            # construct the window handler and run it. when the dialog closes, it will continue by
            # calling open default project.
            def complete_find_existing_projects() -> None:
                self.__open_default_project(profile_dir, is_created)
                self._exit_prevent_close_state()

            self._enter_prevent_close_state()

            window_handler = FindExistingProjectsWindowHandler(completion_fn=complete_find_existing_projects)
            window_handler.run(window, app=self)

            # define an async routine that will perform the finding of the existing projects.
            # this is just a loop that yields via asyncio.sleep periodically. the loop loads
            # the projects and updates the progress and message value models in the dialog.
            # when finished, it asks the window to close on its next periodic call. it is
            # necessary to close this way because the close request may close the event loop
            # in which we're executing. so queueing the close request avoids that.
            async def find_existing_projects() -> None:
                recent_library_paths = self.get_recent_library_paths()
                for index, library_path in enumerate(recent_library_paths):
                    window_handler.progress_value_model.value = 100 * index // len(recent_library_paths)
                    window_handler.message_str_model.value = str(library_path.name)
                    logging.getLogger("loader").info(f"Adding existing project {index + 1}/{len(recent_library_paths)} {library_path}")
                    await asyncio.sleep(0)
                    assert self.__profile
                    self.__profile.add_project_folder(pathlib.Path(library_path), load=False)
                window_handler.progress_value_model.value = 100
                window_handler.message_str_model.value = _("Finished")
                await asyncio.sleep(1)
                window_handler.window.queue_request_close()

            # launch the find existing projects task asynchronously.
            window_handler.window.event_loop.create_task(find_existing_projects())
            return True
        else:
            # continue with opening the default project
            return self.__open_default_project(profile_dir, is_created)

    def __open_default_project(self, profile_dir: typing.Optional[pathlib.Path], is_created: bool) -> bool:
        # if the default project is known, open it.
        # if it fails, ask the user to select another project.
        # if no default project is known, ask the user to choose one.

        profile = self.profile

        project_reference: typing.Optional[Profile.ProjectReference] = None

        if not project_reference:
            project_reference = profile.get_project_reference(profile.last_project_reference) if profile.last_project_reference else None

        update_last_project_reference = True

        keyboard_modifiers = self.ui.get_keyboard_modifiers(True)
        if keyboard_modifiers and keyboard_modifiers.shift:  # pass True since there has been no event yet
            project_reference = None
            update_last_project_reference = False

        if project_reference and project_reference.is_valid:
            try:
                document_controller = self.open_project_window(project_reference, update_last_project_reference)
            except Exception:
                import traceback
                traceback.print_exc()
                self.show_ok_dialog(_("Error Opening Project"), _("Unable to open default project."), completion_fn=self.show_choose_project_dialog)
                return True

            if profile_dir is None:
                # output log message unless we passed a profile_dir for testing.
                logging.getLogger("loader").info("Welcome to Nion Swift.")

            selected_display_panel = document_controller.selected_display_panel
            if selected_display_panel and is_created and len(document_controller.document_model.display_items) > 0:
                selected_display_panel.set_display_panel_display_item(document_controller.document_model.display_items[0])
                selected_display_panel.perform_action("set_fill_mode")
        else:
            self.show_choose_project_dialog()

        return True

    def open_project_manager(self) -> None:
        if not self.is_dialog_type_open(ProjectPanel.ProjectDialog):
            project_dialog = ProjectPanel.ProjectDialog(self.ui, self)
            project_dialog.show()

    def open_project_window(self, project_reference: Profile.ProjectReference, update_last_project_reference: bool = True) -> DocumentController.DocumentController:
        assert self.__profile

        self.__profile.read_project(project_reference)

        document_model = project_reference.document_model
        assert document_model

        document_model.create_default_data_groups()
        document_model.start_dispatcher()

        # create the document controller
        document_controller = self.create_document_controller(document_model, "library", project_reference=project_reference)

        if update_last_project_reference:
            self.__profile.last_project_reference = project_reference.uuid

        project_reference.last_used = datetime.datetime.now()

        return document_controller

    def show_open_project_dialog(self) -> None:
        profile = self.profile
        with self.prevent_close():
            ui = self.ui
            filter_str = "Projects (*.nsproj);;Legacy Libraries (*.nslib);;All Files (*.*)"
            import_dir = ui.get_persistent_string("open_directory", ui.get_document_location())
            paths, selected_filter, selected_directory = ui.get_file_paths_dialog(_("Add Existing Library"), import_dir, filter_str)
            if len(paths) == 1:
                ui.set_persistent_string("open_directory", selected_directory)
                project_reference = profile.open_project(pathlib.Path(paths[0]))
                if project_reference:
                    self.open_project_reference(project_reference)
                else:
                    self.show_ok_dialog(_("Error Opening Project"), _("Unable to open project."), completion_fn=self.show_choose_project_dialog)

    def show_choose_project_dialog(self) -> None:
        with self.prevent_close():
            u = Declarative.DeclarativeUI()
            button_row = u.create_row(u.create_push_button(text=_("New..."), on_clicked="new_project"),
                                      u.create_push_button(text=_("Open..."), on_clicked="open_project"),
                                      u.create_stretch(),
                                      u.create_push_button(text=_("Cancel"), on_clicked="close_window"),
                                      u.create_push_button(text=_("Open Selected"), on_clicked="open_recent"),
                                      spacing=8)

            project_references_model = ListModel.FilteredListModel(container=self.__profile, items_key="project_references")
            project_references_model.filter = ListModel.PredicateFilter(lambda pr: bool(pr.project_state != "loaded"))
            project_references_model.sort_key = lambda pr: pr.recents_key
            project_references_model.sort_reverse = True

            class ProjectReferenceItem:
                # provides a str converter and a tool tip.
                def __init__(self, project_reference: Profile.ProjectReference):
                    self.project_reference = project_reference

                def __str__(self) -> str:
                    project_reference = self.project_reference
                    project_title = project_reference.title or str()
                    project_title += " [" + (project_reference.last_used or project_reference.modified).strftime('%Y-%m-%d') + "]"
                    if project_reference.project_state == "needs_upgrade":
                        project_title += " " + _("(NEEDS UPGRADE)")
                    elif project_reference.project_state == "missing":
                        project_title += " " + _("(MISSING)")
                    elif project_reference.project_state != "unloaded" or project_reference.project_version != FileStorageSystem.PROJECT_VERSION:
                        project_title += " " + _("(UNREADABLE)")
                    return project_title

                @property
                def tool_tip(self) -> str:
                    return str(self.project_reference.path)

            project_reference_items_model = ListModel.MappedListModel(container=project_references_model,
                                                                      master_items_key="project_references",
                                                                      items_key="project_reference_items",
                                                                      map_fn=ProjectReferenceItem)

            item_list = u.create_list_box(items_ref="@binding(list_property_model.value)",
                                          current_index="@binding(current_index)",
                                          min_height=180, min_width=240,
                                          size_policy_horizontal="expanding",
                                          size_policy_vertical="expanding",
                                          on_item_selected="recent_item_selected",
                                          on_item_handle_context_menu="item_handle_context_menu")

            main_column = u.create_column(u.create_label(text=_("Recent Projects")),
                                          item_list,
                                          u.create_spacing(13),
                                          button_row, spacing=8, min_width=360)
            window = u.create_window(main_column, title=_("Choose Project"), margin=12, window_style="tool")

            def open_project_reference(project_reference: Profile.ProjectReference) -> None:
                self.open_project_reference(project_reference)

            def show_open_project_dialog() -> None:
                self.show_open_project_dialog()

            def show_new_project_dialog() -> None:
                NewProjectAction().invoke(UIWindow.ActionContext(self, None, None))

            class ChooseProjectHandler(Declarative.WindowHandler):
                def __init__(self, application: Application):
                    super().__init__()
                    self.__application = application
                    self.current_index = 0
                    self.list_property_model = ListModel.ListPropertyModel(project_reference_items_model)
                    self.__application._enter_prevent_close_state()

                def close(self) -> None:
                    self.__application._exit_prevent_close_state()
                    super().close()

                def recent_item_selected(self, widget: Declarative.UIWidget, current_index: int) -> None:
                    if 0 <= current_index < len(project_reference_items_model.project_reference_items):
                        # to ensure the application does not close upon closing the last window, force it
                        # to stay open while the window is closed and another reopened.
                        with self.__application.prevent_close():
                            self.close_window()
                            project_reference_item = project_reference_items_model.project_reference_items[current_index]
                            open_project_reference(project_reference_item.project_reference)

                def new_project(self, widget: Declarative.UIWidget) -> None:
                    # to ensure the application does not close upon closing the last window, force it
                    # to stay open while the window is closed and another reopened.
                    with self.__application.prevent_close():
                        show_new_project_dialog()
                        self.close_window()

                def open_project(self, widget: Declarative.UIWidget) -> None:
                    # to ensure the application does not close upon closing the last window, force it
                    # to stay open while the window is closed and another reopened.
                    with self.__application.prevent_close():
                        show_open_project_dialog()
                        self.close_window()

                def open_recent(self, widget: Declarative.UIWidget) -> None:
                    self.recent_item_selected(widget, self.current_index)

                def item_handle_context_menu(self, widget: Declarative.UIWidget, *,
                                             gx: int, gy: int,
                                             index: typing.Optional[int], **kwargs: typing.Any) -> bool:
                    if index is not None:
                        project_reference_item = project_reference_items_model.project_reference_items[index]
                        menu = self.window.create_context_menu()
                        menu.add_menu_item(_(f"Open Project Location"), functools.partial(ProjectPanel.reveal_project, project_reference_item.project_reference))
                        menu.add_separator()

                        def remove_project(index: int) -> None:
                            project_reference_item = project_reference_items_model.project_reference_items[index]
                            profile = self.__application.profile
                            profile.remove_project_reference(project_reference_item.project_reference)

                        menu.add_menu_item(_(f"Remove Project from List"), functools.partial(remove_project, index))
                        menu.popup(gx, gy)
                    return True

            ChooseProjectHandler(self).run(window, app=self, persistent_id="choose_project")

    def get_recent_library_paths(self) -> typing.List[pathlib.Path]:
        workspace_history = self.ui.get_persistent_object("workspace_history", list())
        return [pathlib.Path(file_path) for file_path in workspace_history if os.path.exists(file_path)]

    def create_document_controller(self, document_model: DocumentModel.DocumentModel, workspace_id: str,
                                   display_item: typing.Optional[DisplayItem.DisplayItem] = None,
                                   project_reference: typing.Optional[Profile.ProjectReference] = None) -> DocumentController.DocumentController:
        self._set_document_model(document_model)  # required to allow API to find document model
        document_controller = DocumentController.DocumentController(self.ui, document_model, workspace_id=workspace_id,
                                                                    app=self, project_reference=project_reference)
        self.document_model_available_event.fire(document_model)
        # attempt to set data item / group
        if display_item:
            display_panel = document_controller.selected_display_panel
            if display_panel:
                display_panel.set_display_panel_display_item(display_item)
        setattr(document_controller, "_dynamic_recent_project_actions", list())
        # restore the ui state and geometry before show so that it pops up in the right location
        document_controller.restore_ui_state_and_geometry()
        document_controller.show()
        return document_controller

    def _set_profile_for_test(self, profile: typing.Optional[Profile.Profile]) -> None:
        self.__profile = profile

    def __establish_profile(self, profile_path: pathlib.Path) -> typing.Tuple[typing.Optional[Profile.Profile], bool]:
        assert profile_path.is_absolute()  # prevents tests from creating temporary files in test directory
        create_new_profile = not profile_path.exists()
        if create_new_profile:
            logging.getLogger("loader").info(f"Creating new profile {profile_path}")
            profile_json = json.dumps({"version": FileStorageSystem.PROFILE_VERSION, "uuid": str(uuid.uuid4())})
            profile_path.write_text(profile_json, "utf-8")
        else:
            logging.getLogger("loader").info(f"Using existing profile {profile_path}")
        storage_system = FileStorageSystem.make_file_persistent_storage_system(profile_path)
        storage_system.load_properties()
        old_cache_path = profile_path.parent / pathlib.Path(profile_path.stem + " Cache").with_suffix(".nscache")
        if old_cache_path.exists():
            logging.getLogger("loader").info(f"Removing old cache {old_cache_path}")
            old_cache_path.unlink()
        cache_dir_path = profile_path.parent / "Cache"
        cache_dir_path.mkdir(parents=True, exist_ok=True)
        profile = Profile.Profile(storage_system=storage_system, cache_dir_path=cache_dir_path)
        profile.read_profile()
        return profile, create_new_profile

    @property
    def document_controllers(self) -> typing.List[DocumentController.DocumentController]:
        return typing.cast(typing.List[DocumentController.DocumentController], self.windows)

    def register_menu_handler(self, new_menu_handler: typing.Callable[[DocumentController.DocumentController], None]) -> typing.Callable[[DocumentController.DocumentController], None]:
        assert new_menu_handler not in self.__menu_handlers
        self.__menu_handlers.append(new_menu_handler)
        # when a menu handler is registered, let it immediately know about existing menu handlers
        for document_controller in self.windows:
            new_menu_handler(typing.cast(DocumentController.DocumentController, document_controller))
        # return the menu handler so that it can be used to unregister (think: lambda)
        return new_menu_handler

    def unregister_menu_handler(self, menu_handler: typing.Callable[[DocumentController.DocumentController], None]) -> None:
        self.__menu_handlers.remove(menu_handler)

    @property
    def menu_handlers(self) -> typing.Sequence[typing.Callable[[DocumentController.DocumentController], None]]:
        return list(self.__menu_handlers)

    def _menu_about_to_show(self, window: UIWindow.Window, menu: UserInterface.Menu) -> bool:
        if menu.menu_id == "recent_projects":
            self.__about_to_show_recent_projects_menu(window, menu)
            return True
        return super()._menu_about_to_show(window, menu)

    def __about_to_show_recent_projects_menu(self, window: UIWindow.Window, menu: UserInterface.Menu) -> None:
        # recent project actions are stored with the window so they can be deleted.
        if hasattr(window, "_dynamic_recent_project_actions"):
            for recent_project_action in getattr(window, "_dynamic_recent_project_actions"):
                menu.remove_action(recent_project_action)
        setattr(window, "_dynamic_recent_project_actions", list())
        profile = self.profile
        project_references: typing.List[Profile.ProjectReference] = list(filter(lambda pr: pr.project_state != "loaded", profile.project_references))
        project_references = sorted(project_references, key=operator.attrgetter("recents_key"), reverse=True)
        for project_reference in project_references[:20]:
            if project_reference.project_state != "loaded":
                project_title = project_reference.title
                project_title += " [" + (project_reference.last_used or project_reference.modified).strftime('%Y-%m-%d') + "]"
                if project_reference.project_state == "needs_upgrade":
                    project_title += " " + _("(NEEDS UPGRADE)")
                elif project_reference.project_state == "missing":
                    project_title += " " + _("(MISSING)")
                elif project_reference.project_state != "unloaded" or project_reference.project_version != FileStorageSystem.PROJECT_VERSION:
                    project_title += " " + _("(UNREADABLE)")
                action = menu.add_menu_item(project_title, functools.partial(self.open_project_reference, project_reference))
                getattr(window, "_dynamic_recent_project_actions").append(action)

    def _get_recent_project_references(self) -> typing.Sequence[Profile.ProjectReference]:
        profile = self.profile
        project_references: typing.List[Profile.ProjectReference] = list(filter(lambda pr: pr.project_state != "loaded", profile.project_references))
        return sorted(project_references, key=operator.attrgetter("recents_key"), reverse=True)

    def create_project_reference(self, directory: pathlib.Path, project_name: str) -> None:
        profile = self.profile
        project_reference = profile.create_project(directory, project_name)
        if project_reference:
            self.switch_project_reference(project_reference)

    def open_project_reference(self, project_reference: Profile.ProjectReference) -> None:
        with self.prevent_close():
            if project_reference.project_version == FileStorageSystem.PROJECT_VERSION and project_reference.project_state == "unloaded":
                self.switch_project_reference(project_reference)
            elif project_reference.project_state == "needs_upgrade":
                def handle_upgrade(result: bool) -> None:
                    profile = self.profile
                    if result and profile:
                        try:
                            new_project_reference = profile.upgrade(project_reference)
                        except FileExistsError:
                            message = _("Upgraded project already exists.")
                            self.show_ok_dialog(_("Error Upgrading Project"), f"{message}\n{project_reference.path}")
                            logging.getLogger("loader").info(f"Project already exists: {project_reference.path}")
                            new_project_reference = None
                        except Exception as e:
                            self.show_ok_dialog(_("Error Upgrading Project"), _("Unable to upgrade project."))
                            import traceback
                            traceback.print_exc()
                            new_project_reference = None
                        if new_project_reference:
                            self.switch_project_reference(new_project_reference)

                self.show_ok_cancel_dialog(_("Project Needs Upgrade"),
                                           _("This project needs to be upgraded to work with this version."),
                                           ok_text=_("Upgrade"),
                                           completion_fn=handle_upgrade)
            else:
                self.show_ok_dialog(_("Error Opening Project"), _("Unable to open project."), completion_fn=self.show_choose_project_dialog)

    def switch_project_reference(self, project_reference: Profile.ProjectReference) -> None:
        for window in self.windows:
            if isinstance(window, DocumentController.DocumentController):
                window.request_close()
        try:
            self.open_project_window(project_reference)
        except Exception:
            self.show_ok_dialog(_("Error Opening Project"), _("Unable to open project."), completion_fn=self.show_choose_project_dialog)

    @classmethod
    def get_nion_swift_version_info(cls) -> typing.List[typing.Tuple[str, str, str]]:
        info = list()
        for path_str in sys.path:
            if path_str.endswith("site-packages"):
                site_packages_path = pathlib.Path(path_str)
                for package in sorted(path.stem for path in site_packages_path.glob("*dist-info")):
                    m = re.match(r"(.*)-(\d+\.\d+.*\d*)", package)
                    g = m.groups() if m else list()
                    if len(g) >= 2:
                        package_name = g[0]
                        package_version = g[1]
                        if package_name.startswith("nion") or package_name in ("python", "scipy", "numpy", "h5py"):
                            info.append((package, package_name, package_version))
        return info

    def show_about_box(self, parent_window: UIWindow.Window) -> None:
        about_dialog = AboutDialog(self.ui, parent_window, self.version_str)
        about_dialog.show()

    def execute_command_list(
            self,
            commands: typing.Sequence[typing.Tuple[str, typing.Optional[typing.Dict[str, typing.Any]]]],
            *,
            command_delay: typing.Optional[float] = None,
            repeat: typing.Optional[int] = None
    ) -> None:
        """Execute a list of application level commands with optional delay between commands and optional repeat count.

        FOR TESTING ONLY.
        """

        def execute_next_command(
                commands: typing.Sequence[typing.Tuple[str, typing.Optional[typing.Dict[str, typing.Any]]]],
                remaining_commands: typing.MutableSequence[
                    typing.Tuple[str, typing.Optional[typing.Dict[str, typing.Any]]]],
                command_delay: float,
                repeat: int
        ) -> None:
            if remaining_commands:
                next_command = remaining_commands.pop(0)
                self.execute_action(next_command[0], parameters=next_command[1])
                self.event_loop.call_later(command_delay, execute_next_command, commands, remaining_commands, command_delay, repeat)
            elif repeat > 0:
                self.event_loop.call_later(command_delay, execute_next_command, commands, list(commands), command_delay, repeat - 1)

        self.event_loop.call_later(command_delay or 0.0, execute_next_command, commands, list(commands), command_delay or 0.0, repeat or 1)


def get_root_dir() -> str:
    root_dir = os.path.dirname((os.path.dirname(os.path.abspath(__file__))))
    path_ascend_count = 2
    for i in range(path_ascend_count):
        root_dir = os.path.dirname(root_dir)
    return root_dir


class FeaturesPreferencePanel:
    def __init__(self) -> None:
        self.identifier = "nion.swift.features-panel"
        self.label = _("Features")

    def build(self, ui: UserInterface.UserInterface, event_loop: typing.Optional[asyncio.AbstractEventLoop] = None, **kwargs: typing.Any) -> Declarative.DeclarativeWidget:
        u = Declarative.DeclarativeUI()

        class FeatureHandler(Declarative.Handler):
            def __init__(self, item: Feature.Feature) -> None:
                super().__init__()
                self.feature = item
                self.ui_view = u.create_column(
                    u.create_row(
                        u.create_check_box(text="@binding(feature.description)", checked="@binding(feature.enabled)"),
                        u.create_stretch(),
                        spacing=8
                    ),
                    spacing=8
                )

        class Handler(Declarative.Handler):
            def __init__(self) -> None:
                super().__init__()
                self.features = Feature.FeatureManager().features
                self.ui_view = u.create_column(
                    u.create_row(
                        u.create_label(text="Optional Features", width=120),
                        u.create_stretch(),
                        spacing=8
                    ),
                    u.create_divider(orientation="horizontal"),
                    u.create_scroll_area(content=u.create_column(u.create_column(items="features", item_component_id="feature-component", spacing=4), u.create_stretch())),
                    u.create_stretch(),
                    spacing=8
                )

            def create_handler(self, component_id: str, container: typing.Any = None, item: typing.Any = None, **kwargs: typing.Any) -> typing.Optional[Declarative.HandlerLike]:
                # this is called to construct contained declarative component handlers within this handler.
                if component_id == "feature-component":
                    assert container is not None
                    assert item is not None
                    return FeatureHandler(item)
                return None

        return Declarative.DeclarativeWidget(ui, event_loop or asyncio.get_event_loop(), Handler())


PreferencesDialog.PreferencesManager().register_preference_pane(FeaturesPreferencePanel())


class NewProjectAction(UIWindow.Action):
    action_id = "project.new_project"
    action_name = _("New Project...")

    def execute(self, context: UIWindow.ActionContext) -> UIWindow.ActionResult:
        raise NotImplementedError()

    def invoke(self, context: UIWindow.ActionContext) -> UIWindow.ActionResult:
        context = typing.cast(DocumentController.DocumentController.ActionContext, context)

        class NewProjectDialog(Dialog.ActionDialog):

            def __init__(self, ui: UserInterface.UserInterface, app: Application, event_loop: asyncio.AbstractEventLoop, profile: Profile.Profile) -> None:
                super().__init__(ui, title=_("New Project"), app=app, persistent_id="new_project_dialog")

                self.directory = self.ui.get_persistent_string("project_directory", self.ui.get_document_location())

                project_base_name = _("Nion Swift Project") + " " + datetime.datetime.now().strftime("%Y%m%d")
                project_base_index = 0
                project_base_index_str = ""
                while os.path.exists(os.path.join(self.directory, project_base_name + project_base_index_str)):
                    project_base_index += 1
                    project_base_index_str = " " + str(project_base_index)

                self.project_name = project_base_name + project_base_index_str

                def handle_close() -> bool:
                    self.request_close()
                    return True

                def handle_new_and_close() -> bool:
                    app.create_project_reference(pathlib.Path(self.directory), self.__project_name_field.text or "untitled")
                    self.request_close()
                    return True

                def verify_project_name(text: str) -> None:
                    valid = text == Utility.simplify_filename(text)
                    if valid:
                        create_project_button.enabled = True
                        create_project_button.tool_tip = None
                        project_name_status_label.text = None
                    else:
                        create_project_button.enabled = False
                        invalid_chars_str = _("Invalid Characters in Project Name")
                        create_project_button.tool_tip = invalid_chars_str
                        project_name_status_label.text = invalid_chars_str
                        project_name_status_label.text_color = "red"
                column = self.ui.create_column_widget()

                directory_header_row = self.ui.create_row_widget()
                directory_header_row.add_spacing(13)
                directory_header_row.add(self.ui.create_label_widget(_("Projects Folder: "), properties={"font": "bold"}))
                directory_header_row.add_stretch()
                directory_header_row.add_spacing(13)

                show_directory_row = self.ui.create_row_widget()
                show_directory_row.add_spacing(26)
                directory_label = self.ui.create_label_widget(self.directory)
                show_directory_row.add(directory_label)
                show_directory_row.add_stretch()
                show_directory_row.add_spacing(13)

                choose_directory_row = self.ui.create_row_widget()
                choose_directory_row.add_spacing(26)
                choose_directory_button = self.ui.create_push_button_widget(_("Choose..."))
                choose_directory_row.add(choose_directory_button)
                choose_directory_row.add_stretch()
                choose_directory_row.add_spacing(13)

                project_name_header_row = self.ui.create_row_widget()
                project_name_header_row.add_spacing(13)
                project_name_header_row.add(self.ui.create_label_widget(_("Project Name: "), properties={"font": "bold"}))
                project_name_header_row.add_stretch()
                project_name_header_row.add_spacing(13)

                project_name_row = self.ui.create_row_widget()
                project_name_row.add_spacing(26)
                project_name_field = self.ui.create_line_edit_widget(properties={"width": 400})
                project_name_field.text = self.project_name
                project_name_field.on_return_pressed = handle_new_and_close
                project_name_field.on_escape_pressed = handle_close
                project_name_field.on_text_edited = verify_project_name
                project_name_row.add(project_name_field)
                project_name_row.add_stretch()
                project_name_row.add_spacing(13)

                project_name_status_label = self.ui.create_label_widget()
                project_name_status_row = self.ui.create_row_widget()
                project_name_status_row.add_spacing(26)
                project_name_status_row.add(project_name_status_label)
                project_name_status_row.add_stretch()
                project_name_status_row.add_spacing(13)

                column.add_spacing(12)
                column.add(directory_header_row)
                column.add_spacing(8)
                column.add(show_directory_row)
                column.add_spacing(8)
                column.add(choose_directory_row)
                column.add_spacing(16)
                column.add(project_name_header_row)
                column.add_spacing(8)
                column.add(project_name_row)
                column.add_spacing(4)
                column.add(project_name_status_row)
                column.add_stretch()
                column.add_spacing(16)

                def choose() -> None:
                    existing_directory, directory = self.ui.get_existing_directory_dialog(_("Choose Project Directory"), self.directory)
                    if existing_directory:
                        self.directory = existing_directory
                        directory_label.text = self.directory
                        self.ui.set_persistent_string("project_directory", self.directory)

                choose_directory_button.on_clicked = choose

                self.add_button(_("Cancel"), lambda: True)
                create_project_button = self.add_button(_("Create Project"), handle_new_and_close)

                self.content.add(column)

                self.__project_name_field = project_name_field

            def close(self) -> None:
                if self.app:
                    self.app._exit_prevent_close_state()
                super().close()

            def show(self, *, size: typing.Optional[Geometry.IntSize] = None, position: typing.Optional[Geometry.IntPoint] = None) -> None:
                super().show(size=size, position=position)
                self.__project_name_field.focused = True
                self.__project_name_field.select_all()

        application = typing.cast(Application, context.application)
        profile = application.profile
        application._enter_prevent_close_state()
        new_project_dialog = NewProjectDialog(application.ui, application, application.event_loop, profile)
        new_project_dialog.show()

        return UIWindow.ActionResult(UIWindow.ActionStatus.FINISHED)


class OpenProjectAction(UIWindow.Action):
    action_id = "project.open_project_dialog"
    action_name = _("Open Project...")

    def execute(self, context: UIWindow.ActionContext) -> UIWindow.ActionResult:
        raise NotImplementedError()

    def invoke(self, context_: UIWindow.ActionContext) -> UIWindow.ActionResult:
        context = typing.cast(DocumentController.DocumentController.ActionContext, context_)
        application = typing.cast(Application, context.application)
        application.show_open_project_dialog()
        return UIWindow.ActionResult(UIWindow.ActionStatus.FINISHED)


class ChooseProjectAction(UIWindow.Action):
    action_id = "project.choose_project"
    action_name = _("Choose Project...")

    def execute(self, context: UIWindow.ActionContext) -> UIWindow.ActionResult:
        raise NotImplementedError()

    def invoke(self, context_: UIWindow.ActionContext) -> UIWindow.ActionResult:
        context = typing.cast(DocumentController.DocumentController.ActionContext, context_)
        application = typing.cast(Application, context.application)
        application.show_choose_project_dialog()
        return UIWindow.ActionResult(UIWindow.ActionStatus.FINISHED)


class SwitchProjectAction(UIWindow.Action):
    action_id = "project.switch_project"
    action_name = _("Switch Project")

    def execute(self, context: UIWindow.ActionContext) -> UIWindow.ActionResult:
        application = typing.cast(Application, context.application)
        with application.prevent_close():
            project_reference: typing.Optional[Profile.ProjectReference] = None
            if project_uuid := context.parameters.get("project_uuid", typing.cast(typing.Optional[uuid.UUID], None)):
                project_reference = application.profile.get_project_reference_by_project_uuid(project_uuid)
            elif project_path_str := context.parameters.get("project_path", str()):
                project_path = pathlib.Path(project_path_str)
                project_reference = application.profile.get_project_reference_by_path(project_path)
                if not project_reference:
                    project_reference = application.profile.open_project(project_path)
            if project_reference and project_reference.project_version == FileStorageSystem.PROJECT_VERSION and project_reference.project_state == "unloaded":
                application.switch_project_reference(project_reference)
        return UIWindow.ActionResult(UIWindow.ActionStatus.FINISHED)


UIWindow.register_action(NewProjectAction())
UIWindow.register_action(OpenProjectAction())
UIWindow.register_action(ChooseProjectAction())
UIWindow.register_action(SwitchProjectAction())



from __future__ import annotations

# standard libraries
import copy
import typing

# third party libraries

# local libraries
from nion.swift.model import DataItem
from nion.swift.model import DisplayItem
from nion.swift.model import Graphics
from nion.swift.model import Persistence
from nion.swift.model import Schema
from nion.utils import Event

if typing.TYPE_CHECKING:
    from nion.swift.model import Project


class DataStructure(Persistence.PersistentObject):
    entity_types: typing.Dict[str, Schema.EntityType] = dict()
    entity_names: typing.Dict[str, str] = dict()
    entity_package_names: typing.Dict[str, str] = dict()

    # regarding naming: https://en.wikipedia.org/wiki/Passive_data_structure
    def __init__(self, *, structure_type: typing.Optional[str] = None, source: typing.Optional[Persistence.PersistentObject] = None) -> None:
        super().__init__()
        self.__properties: typing.Dict[str, typing.Any] = dict()
        self.__referenced_object_proxies: typing.Dict[str, Persistence.PersistentObjectProxy[Persistence.PersistentObject]] = dict()
        self.define_type("data_structure")
        self.define_property("structure_type", structure_type, changed=self.__structure_type_changed, hidden=True)
        self.define_property("source_specifier", changed=self.__source_specifier_changed, key="source_uuid", hidden=True)
        # properties are handled explicitly
        self.data_structure_changed_event = Event.Event()
        self.data_structure_objects_changed_event = Event.Event()
        self.__source_reference = self.create_item_reference(item=source)
        self.source_specifier = Persistence.write_persistent_specifier(source.uuid) if source else None
        self.__entity: typing.Optional[Schema.Entity] = None
        self.__entity_property_changed_event_listener: typing.Optional[Event.EventListener] = None
        self.__create_entity()

    def close(self) -> None:
        if self.__entity_property_changed_event_listener:
            self.__entity_property_changed_event_listener.close()
            self.__entity_property_changed_event_listener = typing.cast(Event.EventListener, None)
        if self.__entity:
            self.__entity.close()
            self.__entity = None
        for referenced_proxy in self.__referenced_object_proxies.values():
            referenced_proxy.close()
        self.__referenced_object_proxies.clear()
        super().close()

    @property
    def structure_type(self) -> str:
        return typing.cast(str, self._get_persistent_property_value("structure_type"))

    @structure_type.setter
    def structure_type(self, value: str) -> None:
        self._set_persistent_property_value("structure_type", value)

    @property
    def source_specifier(self) -> typing.Optional[Persistence._SpecifierType]:
        return typing.cast(typing.Optional[Persistence._SpecifierType], self._get_persistent_property_value("source_specifier"))

    @source_specifier.setter
    def source_specifier(self, value: typing.Optional[Persistence._SpecifierType]) -> None:
        self._set_persistent_property_value("source_specifier", value)

    @classmethod
    def register_entity(cls, entity_type: Schema.EntityType, *, entity_name: typing.Optional[str] = None, entity_package_name: typing.Optional[str] = None, **kwargs: typing.Any) -> None:
        DataStructure.entity_types[entity_type.entity_id] = entity_type
        if entity_name:
            DataStructure.entity_names[entity_type.entity_id] = entity_name
        if entity_package_name:
            DataStructure.entity_package_names[entity_type.entity_id] = entity_package_name

    @classmethod
    def unregister_entity(cls, entity_type: Schema.EntityType) -> None:
        DataStructure.entity_types.pop(entity_type.entity_id)
        DataStructure.entity_names.pop(entity_type.entity_id, None)
        DataStructure.entity_package_names.pop(entity_type.entity_id, None)

    def __getattr__(self, name: str) -> typing.Any:
        properties = self.__dict__.get("_DataStructure__properties", dict())
        referenced_object_proxies = self.__dict__.get("_DataStructure__referenced_object_proxies", dict())
        if name in referenced_object_proxies:
            return self.get_referenced_object(name)
        if name in properties:
            return properties[name]
        raise AttributeError("%r object has no attribute %r" % (self.__class__, name))

    def __setattr__(self, name: str, value: typing.Any) -> None:
        properties = self.__dict__.get("_DataStructure__properties", dict())
        referenced_object_proxies = self.__dict__.get("_DataStructure__referenced_object_proxies", dict())
        if name in referenced_object_proxies:
            if value is not None:
                self.set_referenced_object(name, value)
            else:
                self.remove_referenced_object(name)
        if name in properties:
            if value is not None:
                self.set_property_value(name, value)
            else:
                self.remove_property_value(name)
        else:
            super().__setattr__(name, value)

    @property
    def project(self) -> Project.Project:
        return typing.cast("Project.Project", self.container)

    def create_proxy(self) -> Persistence.PersistentObjectProxy[DataStructure]:
        return self.project.create_item_proxy(item=self)

    @property
    def item_specifier(self) -> Persistence.PersistentObjectSpecifier:
        return Persistence.PersistentObjectSpecifier(self.uuid)

    def read_from_dict(self, properties: Persistence.PersistentDictType) -> None:
        super().read_from_dict(properties)
        self.__properties = typing.cast(typing.Dict[str, typing.Any], properties.get("properties"))
        for property_name, value in self.__properties.items():
            self.__configure_reference_proxy(property_name, value, None)
        self.__create_entity()

    def __create_entity(self) -> None:
        if self.__entity_property_changed_event_listener:
            self.__entity_property_changed_event_listener.close()
            self.__entity_property_changed_event_listener = typing.cast(Event.EventListener, None)
            if self.__entity:
                self.__entity.close()
            self.__entity = None

        if self.structure_type in DataStructure.entity_types:
            self.__entity = DataStructure.entity_types[self.structure_type].create(typing.cast(Schema.EntityContext, self.persistent_object_context), self.__properties)

            def entity_property_changed(name: str) -> None:
                if name != "type" and self.__entity:
                    value = self.__entity._get_field_value(name)
                    self.__properties[name] = value
                    reference_object_proxy = self.__referenced_object_proxies.pop(name, None)
                    if reference_object_proxy:
                        reference_object_proxy.close()
                    self.__configure_reference_proxy(name, value, None)
                    self.data_structure_changed_event.fire(name)
                    self.property_changed_event.fire(name)
                self._update_persistent_property("properties", self.__properties)

            self.__entity_property_changed_event_listener = self.__entity.property_changed_event.listen(entity_property_changed)

    def __configure_reference_proxy(self, property_name: str, value: typing.Any, item: typing.Optional[Persistence.PersistentObject]) -> None:
        if isinstance(value, dict) and value.get("type") in {"data_item", "display_item", "data_source", "graphic-specifier", "structure"} and "uuid" in value:
            self.__referenced_object_proxies[property_name] = self.create_item_proxy(item_specifier=Persistence.read_persistent_specifier(value["uuid"]), item=item)

    def write_to_dict(self) -> Persistence.PersistentDictType:
        properties = super().write_to_dict()
        properties["properties"] = copy.deepcopy(self.__properties)
        return properties

    def persistent_object_context_changed(self) -> None:
        super().persistent_object_context_changed()
        if self.__entity:
            self.__entity._set_entity_context(typing.cast(Schema.EntityContext, self.persistent_object_context))

    @property
    def source(self) -> typing.Optional[Persistence.PersistentObject]:
        return self.__source_reference.item

    @source.setter
    def source(self, source: typing.Optional[Persistence.PersistentObject]) -> None:
        self.__source_reference.item = source
        self.source_specifier = Persistence.write_persistent_specifier(source.uuid) if source else None

    @property
    def entity(self) -> typing.Optional[Schema.Entity]:
        return self.__entity

    def __structure_type_changed(self, name: str, structure_type: str) -> None:
        self.__create_entity()
        self.property_changed_event.fire("structure_type")

    def __source_specifier_changed(self, name: str, d: Persistence._SpecifierType) -> None:
        self.__source_reference.item_specifier = Persistence.read_persistent_specifier(d)

    def set_property_value(self, property: str, value: typing.Any) -> None:
        self.__properties[property] = value
        reference_object_proxy = self.__referenced_object_proxies.pop(property, None)
        if reference_object_proxy:
            reference_object_proxy.close()
        self.__configure_reference_proxy(property, value, None)
        self.data_structure_changed_event.fire(property)
        self.property_changed_event.fire(property)
        self._update_persistent_property("properties", self.__properties)

    def remove_property_value(self, property: str) -> None:
        if property in self.__properties:
            self.__properties.pop(property)
            reference_object_proxy = self.__referenced_object_proxies.pop(property, None)
            if reference_object_proxy:
                reference_object_proxy.close()
            self.data_structure_changed_event.fire(property)
            self.property_changed_event.fire(property)
            self._update_persistent_property("properties", self.__properties)

    def get_property_value(self, property: str, default_value: typing.Any = None) -> typing.Any:
        return self.__properties.get(property, default_value)

    def set_referenced_object(self, property: str, item: typing.Any) -> None:
        assert item is not None
        if item != self.get_referenced_object(property):
            object_type = "data_item" if isinstance(item, DataItem.DataItem) else None
            self.__properties[property] = get_object_specifier(item, object_type)
            reference_object_proxy = self.__referenced_object_proxies.pop(property, None)
            if reference_object_proxy:
                reference_object_proxy.close()
            self.__configure_reference_proxy(property, self.__properties[property], item)
            self.data_structure_changed_event.fire(property)
            self.property_changed_event.fire(property)
            self._update_persistent_property("properties", self.__properties)
            self.data_structure_objects_changed_event.fire()

    def remove_referenced_object(self, property: str) -> None:
        self.remove_property_value(property)

    def get_referenced_object(self, property: str) -> typing.Any:
        return self.__referenced_object_proxies[property].item if property in self.__referenced_object_proxies else None

    @property
    def referenced_objects(self) -> typing.List[typing.Optional[Persistence.PersistentObject]]:
        return list(referenced_object_proxy.item for referenced_object_proxy in self.__referenced_object_proxies.values())


def get_object_specifier(object: typing.Optional[Persistence.PersistentObject], object_type: typing.Optional[str] = None,
                         project: typing.Optional[Project.Project] = None) -> typing.Optional[Persistence.PersistentDictType]:
    # project is passed for testing only
    if isinstance(object, DataItem.DataItem):
        specifier = Persistence.PersistentObjectSpecifier(object.uuid)
        specifier_uuid = specifier.item_uuid if specifier else object.uuid
        d = {"version": 1, "type": object_type or "data_item", "uuid": str(specifier_uuid)}
        return d
    if object and object_type in ("xdata", "display_xdata", "cropped_xdata", "cropped_display_xdata", "filter_xdata", "filtered_xdata"):
        assert isinstance(object, DisplayItem.DisplayDataChannel)
        specifier = Persistence.PersistentObjectSpecifier(object.uuid)
        specifier_uuid = specifier.item_uuid if specifier else object.uuid
        d = {"version": 1, "type": object_type, "uuid": str(specifier_uuid)}
        return d
    if isinstance(object, DisplayItem.DisplayDataChannel):
        # should be "data_source" but requires file format change
        specifier = Persistence.PersistentObjectSpecifier(object.uuid)
        specifier_uuid = specifier.item_uuid if specifier else object.uuid
        d = {"version": 1, "type": "data_source", "uuid": str(specifier_uuid)}
        return d
    elif isinstance(object, Graphics.Graphic):
        specifier = Persistence.PersistentObjectSpecifier(object.uuid)
        specifier_uuid = specifier.item_uuid if specifier else object.uuid
        d = {"version": 1, "type": "graphic-specifier", "uuid": str(specifier_uuid)}
        return d
    elif isinstance(object, DataStructure):
        specifier = Persistence.PersistentObjectSpecifier(object.uuid)
        specifier_uuid = specifier.item_uuid if specifier else object.uuid
        d = {"version": 1, "type": "structure", "uuid": str(specifier_uuid)}
        return d
    elif isinstance(object, DisplayItem.DisplayItem):
        specifier = Persistence.PersistentObjectSpecifier(object.uuid)
        specifier_uuid = specifier.item_uuid if specifier else object.uuid
        d = {"version": 1, "type": "display_item", "uuid": str(specifier_uuid)}
        return d
    return None
